#!/usr/bin/env python3
# -*- coding: utf-8 -*


import copy
import itertools
import pathlib
import re
import shutil

from typing import Union, List

import omegaconf

from ase import Atoms

from .. import AbstractTrainer, AbstractPotentialManager, DummyCalculator
from .. import CommitteeCalculator


def parse_reann_input_config(para: Union[str, pathlib.Path]) -> List[str]:
    """Parse `para/input_density` and return the `atomtype`.

    Args:
        para: The para directory path.

    Return:
        A List of Strings, which is the type_list.

    """
    type_list = []

    para = pathlib.Path(para)
    p_density = para / "input_density"
    if p_density.exists():
        with open(p_density, "r") as fopen:
            lines = fopen.readlines()
        for line in lines:
            if line.strip().startswith("atomtype"):
                # type_list = line.strip().split("#")[0].split("=")[1]
                m = re.findall("\[.*\]", line)
                assert len(m) == 1
                type_list = [str(x.strip(" '\"")) for x in m[0][1:-1].split(",")]
                break
        else:
            raise RuntimeError(f"No atomtype found in {str(para)}.")
    else:
        raise FileNotFoundError(f"{str(para)} does not exist.")

    return type_list


def _convert_a_single_line(line: str):
    """"""
    # - clean line
    data = line.strip().split("#")[0]
    k, v = data.split("=")
    k = k.strip()
    v = v.strip()

    def is_number(s) -> bool:
        """"""
        try:
            float(s)
            return True
        except ValueError:
            ...

        return False

    def is_list(s) -> bool:
        """"""
        m = re.findall("\[.*\]", s)
        if len(m) == 0:
            return False
        elif len(m) == 1:
            return True
        else:
            raise RuntimeError(f"Fail to parse {s}.")

    # - convert value
    if is_number(v):
        v = v.strip()
        if v.isdigit():
            v = int(v)
        else:
            v = float(v)
    elif is_list(v):
        v_ = []
        entries = v.strip()[1:-1].split(",")
        for x in entries:
            x = x.strip()
            if is_number(x):
                if x.isdigit():
                    x = int(x)
                else:
                    x = float(x)
            else:  # assume it is a simple string
                x = x.strip(" '\"")
            v_.append(x)
        v = v_
    else:  # assume it is a simple string
        v = v.strip(" '\"")

    return (k, v)


def load_reann_input_para(para: Union[str, pathlib.Path]) -> dict:
    """"""
    params = {}

    para = pathlib.Path(para)

    # - input_density
    with open(para / "input_density", "r") as fopen:
        lines = fopen.readlines()

    params["density"] = {}
    for line in lines:
        if not line.strip().startswith("#"):
            k, v = _convert_a_single_line(line)
            params["density"][k] = v

    # - input_dnn
    with open(para / "input_nn", "r") as fopen:
        lines = fopen.readlines()

    params["nn"] = {}
    for line in lines:
        if not line.strip().startswith("#"):
            k, v = _convert_a_single_line(line)
            params["nn"][k] = v

    return params

def dump_reann_input_para(config_params: dict, para: Union[str, pathlib.Path]):
    """"""
    with open(para/"input_density", "w") as fopen:
        content = ""
        for k, v in config_params["density"].items():
            if isinstance(v, str) and (v != "True" or v != "False"):
                content += "{}=\"{}\"\n".format(k, str(v))
            else:
                content += f"{k}={str(v)}\n"
        fopen.write(content)

    with open(para/"input_nn", "w") as fopen:
        content = ""
        for k, v in config_params["nn"].items():
            if isinstance(v, str) and (v != "True" and v != "False"):
                content += "{}=\"{}\"\n".format(k, str(v))
            else:
                content += f"{k}={str(v)}\n"
        fopen.write(content)

    return


def _atoms2reannconfig(atoms: Atoms, point: int) -> str:
    """"""
    lattice_constants = atoms.get_cell(complete=True).flatten().tolist()

    content = f"{point = }\n"
    content += (("{:>12.8f}  " * 3 + "\n") * 3).format(*lattice_constants)
    content += ("pbc " + "{:<d} " * 3 + "\n").format(*atoms.get_pbc())

    # TODO: If input atoms have no forces?
    for s, m, pos, frc in zip(
        atoms.get_chemical_symbols(),
        atoms.get_masses(),
        atoms.get_positions(),
        atoms.get_forces(),
    ):
        content += (
            f"{s:<3s}  {m:<8.4f}  "
            + f"{pos[0]:>24.8f} {pos[1]:>24.8f} {pos[2]:>24.8f}"
            + f"{frc[0]:>24.8f} {frc[1]:>24.8f} {frc[2]:>24.8f}"
            + "\n"
        )

    # TODO: use electronic free energy?
    content += f"abprop: {atoms.get_potential_energy()}\n"

    return content


def convert_structures_to_reannconfig(
    structures: List[Atoms], fpath: Union[str, pathlib.Path]
) -> None:
    """"""
    content = ""
    for i, atoms in enumerate(structures):
        content += _atoms2reannconfig(atoms, i + 1)

    fpath.parent.mkdir(parents=True, exist_ok=True)
    with open(fpath, "w") as fopen:
        fopen.write(content)

    return content


class ReannDataloader:

    name: str = "reann"

    def __init__(self, batchsize: int, directory: Union[str, pathlib.Path]="./", *args, **kwargs) -> None:
        """"""
        self.batchsize = batchsize
        self.directory = pathlib.Path(directory).resolve()

        return

    def as_dict(
        self,
    ) -> dict:
        """"""
        params = {}
        params["name"] = self.name
        params["batchsize"] = self.batchsize
        params["directory"] = str(self.directory.resolve())

        return params


class ReannTrainer(AbstractTrainer):

    name = "reann"
    command = ""
    freeze_command = ""
    prefix = "config"

    def __init__(
        self,
        config: Union[str, pathlib.Path],
        type_list: List[str] = None,
        train_epochs: int = 200,
        print_epochs: int = 5,
        directory=".",
        command="train",
        freeze_command="freeze",
        random_seed: Union[int, dict] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            config,
            type_list,
            train_epochs,
            print_epochs,
            directory,
            command,
            freeze_command,
            random_seed,
            *args,
            **kwargs,
        )

        # self.config = pathlib.Path(self.config).resolve()
        if isinstance(config, dict) or isinstance(config, omegaconf.dictconfig.DictConfig):
            self.config = config
        elif isinstance(config, str) or isinstance(config, pathlib.Path):
            self.config = load_reann_input_para(config)
        else:
            raise RuntimeError(f"Fail to parse {config = }.")

        self._type_list = self.config["density"]["atomtype"]

        return

    @property
    def frozen_name(self):
        """"""
        return f"PES.pt"
    
    @property
    def ckpt_name(self) -> str:
        """"""

        return "REANN.pth"

    def _resolve_train_command(self, *args, **kwargs):
        """"""

        return self.command

    def _resolve_freeze_command(self, *args, **kwargs):
        """"""

        return self.freeze_command
    
    def get_checkpoint(self):
        """"""

        return pathlib.Path(self.directory/self.ckpt_name).resolve()
    
    def _train_from_the_restart(self, dataset, init_model) -> str:
        """Train from the restart"""
        def _train_from_the_scratch(dataset, init_model) -> str:
            # NOTE: `para/input*` will be overwritten by current self.config
            command = self._train_from_the_scratch(dataset, init_model)
            if init_model is not None:
                self._print(f"{self.name} init training from model {init_model}.")
                (self.directory/self.ckpt_name).unlink(missing_ok=True)
                shutil.copyfile(init_model, self.directory/self.ckpt_name)
                prev_config = load_reann_input_para(self.directory/"para")
                prev_config["nn"]["table_init"] = 1
                _ = dump_reann_input_para(prev_config, self.directory/"para")
            else:
                ...
            
            return command

        if not self.directory.exists():
            command = _train_from_the_scratch(dataset, init_model)
        else:
            ckpt_info = self.directory/self.ckpt_name
            if ckpt_info.exists() and ckpt_info.stat().st_size != 0:
                # -
                log_path = self.directory/"nn.err"
                if log_path.exists():
                    with open(self.directory/"nn.err", "r") as fopen:
                        lines = fopen.readlines()
                    epoch_lines = [l for l in lines if l.strip().startswith("Epoch")]
                    try:
                        end_epoch = int(epoch_lines[-1].split()[1])
                        self._debug(f"{end_epoch =}")
                    except:
                        end_epoch = 0
                        self._print(f"The endline of `nn.err` is strange.")
                    # - 
                    prev_config = load_reann_input_para(self.directory/"para")
                    prev_config["nn"]["table_init"] = 1
                    prev_config["nn"]["Epoch"] = prev_config["nn"]["Epoch"] - end_epoch
                    prev_config["nn"]["patience_epoch"] = 0
                    assert prev_config["nn"]["Epoch"] >= 0
                    _ = dump_reann_input_para(prev_config, self.directory/"para")
                    self._print(f"{self.name} restarts training from epoch {end_epoch}.")
                    command = self._resolve_train_command()
                else:
                    # NOTE: This is a new training but init from a previous model.
                    command = _train_from_the_scratch(dataset, init_model)
            else:
                # restart from the scratch and overwrite exists.
                command = _train_from_the_scratch(dataset, init_model)

        return command

    def _prepare_dataset(self, dataset, *args, **kwargs):
        """Prepare a reann dataset for training.

        Currently, it only supports converting xyz dataset.

        """
        self._print(f"{dataset = }")
        # NOTE: make sure the dataset path exists, sometimes it will be access
        #       before training to create a shared dataset
        self.directory.mkdir(parents=True, exist_ok=True)

        if not isinstance(dataset, ReannDataloader):
            set_names, train_frames, test_frames, adjusted_batchsizes = (
                dataset.split_train_and_test()
            )

            # NOTE: reann does not support split-system training,
            #       so we need merge all structures into one List
            train_frames = itertools.chain(*train_frames)
            convert_structures_to_reannconfig(
                train_frames, self.directory / "train" / "configuration"
            )

            test_frames = itertools.chain(*test_frames)
            convert_structures_to_reannconfig(
                test_frames, self.directory / "val" / "configuration"
            )

            dataset = ReannDataloader(
                directory=self.directory, batchsize=dataset.batchsize
            )
        else:
            ...

        return dataset

    def write_input(self, dataset, *args, **kwargs):
        """"""
        self._print(f"write {self.name} inputs...")

        # - convert dataset
        dataset = self._prepare_dataset(dataset)

        # - copy input file
        (self.directory / "para").mkdir(parents=True, exist_ok=True)

        config_params = copy.deepcopy(self.config)
        config_params["nn"]["Epoch"] = self.train_epochs
        config_params["nn"]["print_epoch"] = self.print_epochs
        config_params["nn"]["batchsize_train"] = dataset.batchsize
        config_params["nn"]["batchsize_val"] = dataset.batchsize
        config_params["nn"]["folder"] = str(dataset.directory.resolve()) + "/"

        dump_reann_input_para(config_params, self.directory/"para")

        return

    def read_convergence(self) -> bool:
        """"""
        self._print(f"check {self.name} training convergence...")
        converged = False

        log_path = self.directory / "nn.err"
        if log_path.exists():
            with open(log_path, "r") as fopen:
                lines = fopen.readlines()
            try:
                end_info = lines[-1]
                if end_info.strip() == "terminated normal":
                    converged = True
                self._debug(f"{end_info = }")
            except:
                self._print(f"The endline of `nn.err` is strange.")
        else:
            ...

        return converged


class ReannManager(AbstractPotentialManager):

    name = "reann"
    implemented_backends = [
        "ase",
    ]

    valid_combinations = (("ase", "ase"),)

    def register_calculator(self, calc_params, *agrs, **kwargs):
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)

        # - some shared params
        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", pathlib.Path.cwd())

        # -
        type_list = calc_params.pop("type_list", [])

        # --- model files
        model_ = calc_params.get("model", [])
        if not isinstance(model_, list):
            model_ = [model_]

        models = []
        for m in model_:
            m = pathlib.Path(m).resolve()
            if not m.exists():
                raise FileNotFoundError(f"Cant find model file {str(m)}")
            models.append(str(m))
        self.calc_params.update(model=models)

        precision = calc_params.pop("precision", "float32")
        assert precision in ["float32", "float64"]

        # TODO: make this a dataclass??
        #       currently, default disable uncertainty estimation
        estimate_uncertainty = calc_params.get("estimate_uncertainty", False)

        # - misc
        max_nneigh = calc_params.get("max_nneigh", 25000)

        # - parse calc_params
        calc = DummyCalculator()
        if self.calc_backend == "ase":
            try:
                import torch
                from .calculators.reann import REANN

                device = torch.device(
                    "cuda" if torch.cuda.is_available() else torch.device("cpu")
                )
                if precision == "float32":
                    precision = torch.float32
                elif precision == "float64":
                    precision = torch.float64
                else:
                    ...
            except:
                raise ModuleNotFoundError(
                    "Please install reann and torch to use the ase interface."
                )

            calcs = []
            for m in models:
                calc = REANN(
                    atomtype=type_list,
                    maxneigh=max_nneigh,
                    getneigh=None,
                    nn=m,
                    device=device,
                    dtype=precision,
                )
                calcs.append(calc)
            if len(calcs) == 1:
                calc = calcs[0]
            elif len(calcs) > 1:
                if estimate_uncertainty:
                    calc = CommitteeCalculator(calcs=calcs)
                else:
                    calc = calcs[0]
            else:
                ...
        else:
            ...

        self.calc = calc

        return

    def switch_uncertainty_estimation(self, status: bool = True):
        """Switch on/off the uncertainty estimation."""
        # NOTE: Sometimes the manager loads several models and supports uncertainty
        #       by committee but the user disables it. We need change the calc to
        #       the correct one as the loaded one is just a single calculator.
        if not hasattr(self, "calc"):
            raise RuntimeError(
                "Fail to switch uncertainty status as it does not have a calc."
            )
        # print(f"{self.calc}")

        # NOTE: make sure manager.as_dict() can have correct param
        self.calc_params["estimate_uncertainty"] = status

        # - convert calculator
        if self.calc_backend == "ase":
            if status:
                if isinstance(self.calc, CommitteeCalculator):
                    ...  # nothing to do
                else:  # reload models
                    self.register_calculator(self.calc_params)
            else:
                if isinstance(self.calc, CommitteeCalculator):
                    # TODO: save previous calc?
                    self.calc = self.calc.calcs[0]
                else:
                    ...
        elif self.calc_backend == "lammps":
            ...
        else:
            # TODO:
            # Other backends cannot have uncertainty estimation,
            # give a warning?
            ...

        return

    def remove_loaded_models(self, *args, **kwargs):
        """Loaded TF models should be removed before any copy.deepcopy operations."""
        self.calc.reset()
        if self.calc_backend == "ase":
            if isinstance(self.calc, CommitteeCalculator):
                for c in self.calc.calcs:
                    c.pes = None
                    c.getneigh = None
            else:
                self.calc.pes = None
                self.calc.getneigh = None
        else:
            ...

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
from typing import NoReturn, Optional, Union, List

from ase import Atoms
from ase.io import read, write
from ase.constraints import FixAtoms

from .builder import StructureBuilder

def read_xsd2(fd) -> Atoms:
    """read xsd file by Material Studio

    NOTE:
        support reading constraints...

    """
    import xml.etree.ElementTree as ET
    from xml.dom import minidom 

    tree = ET.parse(fd)
    root = tree.getroot()

    atomtreeroot = root.find("AtomisticTreeRoot")
    # if periodic system
    if atomtreeroot.find("SymmetrySystem") is not None:
        symmetrysystem = atomtreeroot.find("SymmetrySystem")
        mappingset = symmetrysystem.find("MappingSet")
        mappingfamily = mappingset.find("MappingFamily")
        system = mappingfamily.find("IdentityMapping")

        coords = list()
        cell = list()
        formula = str()

        names = list()
        restrictions = list() 

        for atom in system:
            if atom.tag == "Atom3d":
                symbol = atom.get("Components")
                formula += symbol

                xyz = atom.get("XYZ")
                if xyz:
                    coord = [float(coord) for coord in xyz.split(",")]
                else:
                    coord = [0.0, 0.0, 0.0]
                coords.append(coord)

                name = atom.get("Name") 
                if name:
                    pass # find name 
                else: 
                    name = symbol + str(len(names)+1) # None due to copy atom 
                names.append(name)

                restriction = atom.get("RestrictedProperties", None)
                if restriction:
                    if restriction.startswith("FractionalXYZ"):  # TODO: may have 1-3 flags
                        restrictions.append(True)
                    else: 
                        raise ValueError("unknown RestrictedProperties")
                else: 
                    restrictions.append(False)
            elif atom.tag == "SpaceGroup":
                avec = [float(vec) for vec in atom.get('AVector').split(',')]
                bvec = [float(vec) for vec in atom.get('BVector').split(',')]
                cvec = [float(vec) for vec in atom.get('CVector').split(',')]

                cell.append(avec)
                cell.append(bvec)
                cell.append(cvec)

        atoms = Atoms(formula, cell=cell, pbc=True)
        atoms.set_scaled_positions(coords)

        # add constraints 
        fixed_indices = [idx for idx, val in enumerate(restrictions) if val]
        if fixed_indices:
            atoms.set_constraint(FixAtoms(indices=fixed_indices))

        # add two atoms constrained optimisation 
        constrained_indices = [
            idx for idx, name in enumerate(names) if name.endswith('_c') 
        ]
        if constrained_indices:
            assert len(constrained_indices) == 2
            atoms.info["copt"] = constrained_indices

        return atoms
        # if non-periodic system
    elif atomtreeroot.find("Molecule") is not None:
        system = atomtreeroot.find("Molecule")

        coords = list()
        formula = str()

        for atom in system:
            if atom.tag == "Atom3d":
                symbol = atom.get("Components")
                formula += symbol

                xyz = atom.get("XYZ")
                coord = [float(coord) for coord in xyz.split(",")]
                coords.append(coord)

        atoms = Atoms(formula, pbc=False)
        atoms.set_scaled_positions(coords)
        return atoms

class ReadBuilder(StructureBuilder):

    def __init__(
        self, fname, index=":", format=None, use_tags=False, 
        directory="./", random_seed=None, *args, **kwargs
    ):
        """"""
        super().__init__(use_tags=use_tags, directory=directory, random_seed=random_seed, *args, **kwargs)

        self.fname = pathlib.Path(fname)
        self.index = index
        self.format = format
        #self.kwargs = kwargs

        return
    
    def run(self, *args, **kwargs):
        """"""
        frames = read(self.fname, self.index, self.format)

        return frames
    
    def as_dict(self) -> dict:
        """"""
        params = {}
        params["method"] = "reader"
        params["fname"] = str(self.fname.resolve())
        params["index"] = self.index
        params["format"] = self.format

        return params


class DirectBuilder(StructureBuilder):
    """This generator directly returns structures that it stores.
    """

    #: Builder's name.
    name: str = "direct"

    default_parameters: dict= {

    }

    #: Stored structures.
    _frames: Optional[List[Atoms]] = None

    #: The file path of stored structures.
    _fpath: Optional[Union[str,pathlib.Path]] = None

    #: Selected structure indices.
    _indices: Union[str,List[int]] = None

    def __init__(
        self, frames: Union[str,pathlib.Path,List[Atoms]], 
        indices: Union[str,List[int]] = None, directory: Union[str,pathlib.Path]="./", 
        *args, **kwargs
    ):
        """Create a direct generator.

        Args:
            frames: The structure file path or a list of ase atoms.
            directory: Working directory.

        """
        super().__init__(directory, *args, **kwargs)

        if isinstance(frames, (str,pathlib.Path)):
            self._fpath = pathlib.Path(frames).resolve()
        else:
            assert all(isinstance(x,Atoms) for x in frames), "Input should be a list of atoms."
            self._frames = frames

        self._indices = indices

        return
    
    @property
    def fpath(self) -> Union[str,pathlib.Path]:
        """Return the file path of stored structures."""
        return self._fpath
    
    @property
    def indices(self) -> Union[str,List[int]]:
        """Return selected indices."""
        return self._indices
    
    def run(self, indices: Union[str,List[int]]=[], *args, **kwargs) -> List[Atoms]:
        """Return stored structures.

        Args:
            indices: Selected frames.

        """
        # - check indices
        if indices:
            indices_ = indices
        else:
            indices_ = self.indices

        assert (bool(self._frames) ^ bool(self.fpath)), "Cant have frames and fpath at the same time."

        # - read frames if it is a path
        if self.fpath:
            # NOTE: custom read_xsd can retrieve stored constraints
            fpath_ = str(self.fpath)
            if fpath_.endswith(".xsd"):
                frames_ = read_xsd2(fpath_)
            else:
                frames_ = read(fpath_, ":")

            # - check whether its a single Atoms object
            if isinstance(frames_, Atoms):
                frames_ = [frames_]
        else:
            if self._frames:
                frames_ = self._frames
            else:
                # NOTE: should go to here
                pass

        # - get structures
        if indices_:
            ret_frames = [frames_[i] for i in indices_]
        else:
            ret_frames = frames_
        return ret_frames
    
    def as_dict(self) -> dict:
        """Return generator parameters"""
        params = dict(
            method = "direct",
            frames = str(self._fpath.resolve()), # TODO: if not exists, and only have _frames
            indices = self.indices
        )

        return params


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*


import copy
import pathlib
import warnings

import numpy as np

from ..manager import AbstractPotentialManager


class GridCalculator():

    directory = "./"

    def __init__(self, dpath, unit="eV", *args, **kwargs) -> None:
        """"""
        data = np.loadtxt(dpath) # Need 3 columns, x, y and z.
        x, y, z = self._preprocess_data(data) # C-order 2D data

        self._x, self._y = x, y
        self._z = z / 96.485
        self._points = np.vstack([x.flatten(), y.flatten()]).T
        self._gradients = np.gradient(z, np.unique(data[:, 0]), np.unique(data[:, 1]))

        return
    
    def _preprocess_data(self, data):
        """"""
        nx = np.size(np.unique(data[:, 0]))
        ny = np.size(np.unique(data[:, 1]))

        # Some people give input where the first column varies fastest. 
        # That is Fortran ordering, and you are liable to get things confused 
        # if you don't take this into account.
        order = "C"
        if data[0, 0] != data[1, 0]:
            order = "F"
        x = data[:, 0].reshape(nx, ny, order=order)
        y = data[:, 1].reshape(nx, ny, order=order)
        z = data[:, 2].reshape(nx, ny, order=order)
        #print(f"x: {x}")
        #print(f"y: {y}")

        # - basic input sanity check
        xdiff = np.diff(x, axis=0)
        ydiff = np.diff(y, axis=1)
        if (not np.all(np.abs(xdiff - xdiff[0]) < 1e-8)) or (not np.all(np.abs(ydiff - ydiff[0]) < 1e-8)):
            warnings.warn(
                "WARNING! The input data is not coming from an equally spaced grid. imshow will be subtly wrong as a result, as each pixel is assumed to represent the same area.",
                UserWarning
            )
        
        return x, y, z
    
    def reset(self):
        """"""

        return
    
    def get_potential_energy(self):
        """"""

        return
    
    def get_forces(self):
        """Negative gradients."""

        return


class GridManager(AbstractPotentialManager):

    name = "grid"

    implemented_backends = ["grid"]
    valid_combinations = (
        ("grid", "grid"),
    )

    """"""

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__()

        return
    
    def register_calculator(self, calc_params, *agrs, **kwargs):
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)

        calc_params = copy.deepcopy(calc_params)

        dpath = calc_params.get("data", None) # Must exist!
        dpath = pathlib.Path(dpath).resolve()
        self.calc_params["data"] = str(dpath)

        self.calc = GridCalculator(dpath=dpath)

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
import pathlib
import time
from typing import Optional, List, Tuple, Union

import omegaconf

import numpy as np

from ase import Atoms
from ase.io import read, write

from .. import config
from ..core.operation import Operation
from ..core.register import registers
from ..core.variable import Variable
from ..data.array import AtomsNDArray
from ..selector.scf import ScfSelector
from ..utils.command import CustomTimer
from ..worker.drive import (
    CommandDriverBasedWorker,
    DriverBasedWorker,
    QueueDriverBasedWorker,
)
from ..worker.interface import ComputerVariable

# --- variable ---


@registers.variable.register
class DriverVariable(Variable):

    def __init__(self, **kwargs):
        """"""
        # - compat
        copied_params = copy.deepcopy(kwargs)
        merged_params = dict(
            # task=copied_params.get("task", "min"),
            backend=copied_params.get("backend", "external"),
            ignore_convergence=copied_params.get("ignore_convergence", False),
        )
        merged_params.update(**copied_params.get("init", {}))
        merged_params.update(**copied_params.get("run", {}))

        initial_value = self._broadcast_drivers(merged_params)

        super().__init__(initial_value)

        return

    def _broadcast_drivers(self, params: dict) -> List[dict]:
        """Broadcast parameters if there were any parameter is a list."""
        # - find longest params
        plengths = []
        for k, v in params.items():
            if isinstance(v, list):
                n = len(v)
            else:  # int, float, string
                n = 1
            plengths.append((k, n))
        plengths = sorted(plengths, key=lambda x: x[1])
        # NOTE: check only has one list params
        assert sum([p[1] > 1 for p in plengths]) <= 1, "only accept one param as list."

        # - convert to dataclass
        params_list = []
        maxname, maxlength = plengths[-1]
        for i in range(maxlength):
            curr_params = {}
            for k, n in plengths:
                if n > 1:
                    v = params[k][i]
                else:
                    v = params[k]
                curr_params[k] = v
            params_list.append(curr_params)

        return params_list


# --- operation ---


def extract_results_from_workers(
    directory: pathlib.Path,
    workers: List[DriverBasedWorker],
    *,
    safe_inspect: bool = True,
    use_archive: bool = True,
    print_func=print,
    debug_func=print,
) -> Tuple[str, AtomsNDArray]:
    """"""
    _print = print_func
    _debug = debug_func

    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=True)

    nworkers = len(workers)
    worker_status = [False] * nworkers

    _debug(f"workers: {workers}")

    trajectories = []
    for i, worker in enumerate(workers):
        # TODO: How to save trajectories into one file?
        #       probably use override function for read/write
        #       i - worker, j - cand
        _print(f"worker: {str(worker.directory)}")
        cached_trajs_dpath = directory / f"{worker.directory.parent.name}-w{i}"
        if not cached_trajs_dpath.exists():
            if safe_inspect:
                # inspect again for using extract without drive
                worker.inspect(resubmit=False)
                if not (worker.get_number_of_running_jobs() == 0):
                    _print(f"{worker.directory} is not finished.")
                    break
            else:
                # If compute enables extract, it has already done the inspects
                # thus we can skip them here.
                ...
            cached_trajs_dpath.mkdir(parents=True, exist_ok=True)
            curr_trajectories = worker.retrieve(
                include_retrieved=True, use_archive=use_archive
            )
            AtomsNDArray(curr_trajectories).save_file(cached_trajs_dpath / "dataset.h5")
        else:
            curr_trajectories = AtomsNDArray.from_file(
                cached_trajs_dpath / "dataset.h5"
            ).tolist()

        trajectories.append(curr_trajectories)

        worker_status[i] = True

    trajectories = AtomsNDArray(trajectories)

    status = "unfinished"
    if all(worker_status):
        status = "finished"
    else:
        ...

    return status, trajectories


def get_shape_data(shape_dir: Union[str, pathlib.Path]):
    # - load previous structures' shape
    if shape_dir.exists():
        inp_shape = np.loadtxt(shape_dir / "shape.dat", dtype=int)
        inp_markers = np.loadtxt(shape_dir / "markers.dat", dtype=int)
    else:
        inp_shape = None
        inp_markers = None

    return inp_shape, inp_markers


def convert_results_to_structures(
    structures: List[Atoms],
    inp_shape,
    inp_markers,
    *,
    reduce_single_worker: bool = True,
    merge_workers: bool = False,
    print_func=print,
    debug_func=print,
):
    """Convert `compute` resulst into structures with a proper shape."""
    # -
    _print = print_func
    _debug = debug_func

    # TODO: Convert to correct input data shape for spc workers...
    #       Optimise the codes here?
    _print(f"target structure shape: {inp_shape}")
    _print(f"input  structure shape: {structures.shape}")
    if inp_shape is not None and inp_markers is not None:
        converted_structures = []
        for curr_structures in structures:  # shape (nworkers, ncandidates, 1)
            curr_structures = list(itertools.chain(*curr_structures))
            inp_shape_ = np.max(inp_markers, axis=0) + 1
            # assert np.allclose(
            #    inp_shape_, inp_shape
            # ), "Inconsistent shape {inp_shape_} vs. {inp_shape}"
            _print(f"previous  structure shape: {inp_shape}")
            _print(f"target    structure shape: {inp_shape_}")
            # - get a full list of indices and fill None to a flatten Atoms List
            # _print(inp_markers)
            curr_converted_structures = []
            full_list = list(itertools.product(*[range(x) for x in inp_shape_]))
            for i, iloc in enumerate(full_list):
                if iloc in inp_markers:
                    curr_converted_structures.append(
                        curr_structures[inp_markers.index(iloc)]
                    )
                else:
                    curr_converted_structures.append(None)
            # _print(len(curr_converted_structures))
            # _print(curr_converted_structures[0])
            # - reshape
            for s in inp_shape_[:0:-1]:
                npoints = len(curr_converted_structures)
                repeats = int(npoints / s)
                reshaped_converted_structures = [
                    [curr_converted_structures[i] for i in range(r * s, (r + 1) * s)]
                    for r in range(repeats)
                ]
                curr_converted_structures = reshaped_converted_structures
            converted_structures.append(curr_converted_structures)
        converted_structures = AtomsNDArray(converted_structures)
    else:  # No data available to convert structures
        converted_structures = structures

    # --- further convert
    if reduce_single_worker:  #  nworkers == 1
        converted_structures = converted_structures[0]

    if merge_workers:  # squeeze the dimension one...
        converted_structures = list(itertools.chain(*converted_structures))
    converted_structures = AtomsNDArray(converted_structures)
    _print(f"extracted_results: {converted_structures}")

    return converted_structures


@registers.operation.register
class compute(Operation):
    """Drive structures."""

    def __init__(
        self,
        builder: Variable,
        worker: Variable,
        batchsize: Optional[int] = None,
        share_wdir: bool = False,
        retain_info: bool = False,
        extract_data: bool = True,
        use_archive: bool = True,
        merge_workers: bool = False,
        reduce_single_worker: bool = True,
        check_scf_convergence: bool = False,
        directory: Union[str, pathlib.Path] = "./",
    ):
        """Initialise a compute operation.

        Args:
            builder: A builder node.
            worker: A worker node.
            batchsize: Worker's batchsize can be overwritten by this.
            share_wdir: Worker's share_wdir can be overwritten by this.
            retain_info: Worker's retain_info parameter.
            directory: Operation's directory that stores results can be set by Session.

        """
        super().__init__(input_nodes=[builder, worker], directory=directory)

        # - worker-run-related settings
        self.batchsize = batchsize
        self.share_wdir = share_wdir
        self.retain_info = retain_info

        # - worker-retrieve-related settings
        self.extract_data = extract_data
        self.use_archive = use_archive
        self.merge_workers = merge_workers
        self.reduce_single_worker = reduce_single_worker

        # -
        self.check_scf_convergence = check_scf_convergence

        return

    def _preprocess_input_nodes(self, input_nodes):
        """"""
        builder, worker = input_nodes
        if isinstance(worker, dict) or isinstance(
            worker, omegaconf.dictconfig.DictConfig
        ):
            worker = ComputerVariable(**worker)

        return builder, worker

    def forward(
        self,
        structures: Union[List[Atoms], AtomsNDArray],
        workers: List[DriverBasedWorker],
    ) -> Union[List[DriverBasedWorker], List[AtomsNDArray]]:
        """Run simulations with given structures and workers.

        Workers' working directory and batchsize are probably set.
        Sometimes we need workers as some operations can be applied while
        other times we need directly the trajectories to save operation definitions
        in our session configuration files.

        Returns:
            Workers with correct directory and batchsize or AtomsNDArray.

        """
        super().forward()

        nworkers = len(workers)

        # TODO: It is better to move this part to driver...
        #       We only convert spc worker structures shape here...
        driver0_dict = workers[0].driver.as_dict()
        if (
            nworkers == 1
            and driver0_dict.get("task", "min") == "min"
            and (driver0_dict.get("steps", 0) <= 0)
        ):  # TODO: spc?
            # - check input data type
            inp_shape, inp_markers = None, None
            if isinstance(structures, AtomsNDArray):
                frames = structures.get_marked_structures()
                nframes = len(frames)
                inp_shape = structures.shape
                inp_markers = [tuple(iloc.tolist()) for iloc in structures.markers]
            else:  # assume it is just a List of Atoms
                frames = structures
                nframes = len(frames)
                inp_shape = (1, nframes)
                inp_markers = [(0, i) for i in range(nframes)]
            # self._print(inp_markers)
            # -- Dump shape data
            # NOTE: Since all workers use the same input structures,
            #       we only need to dump once here
            shape_dir = self.directory / "_shape"
            shape_dir.mkdir(parents=True, exist_ok=True)
            np.savetxt(
                shape_dir / "shape.dat", np.array(inp_shape, dtype=np.int32), fmt="%8d"
            )
            np.savetxt(
                shape_dir / "markers.dat",
                np.array(inp_markers, dtype=np.int32),
                fmt="%8d",
            )
        else:
            if isinstance(structures, AtomsNDArray):
                frames = structures.get_marked_structures()
            else:  # assume it is just a plain List of Atoms
                frames = structures
            nframes = len(frames)
            inp_shape, inp_markers = None, None

        # - create workers
        for i, worker in enumerate(workers):
            worker.directory = self.directory / f"w{i}"
            if self.batchsize is not None:
                worker.batchsize = self.batchsize
            else:
                worker.batchsize = nframes
            # if self.share_wdir and worker.scheduler.name == "local":
            if self.share_wdir:
                worker._share_wdir = True
            if self.retain_info:
                worker._retain_info = True
        nworkers = len(workers)

        if nworkers == 1:
            workers[0].directory = self.directory

        # - run workers
        worker_status = []
        for i, worker in enumerate(workers):
            flag_fpath = worker.directory / "FINISHED"
            self._print(f"run worker {i} for {nframes} nframes")
            if not flag_fpath.exists():
                worker.run(frames)
                worker.inspect(resubmit=True)  # if not running, resubmit
                if worker.get_number_of_running_jobs() == 0:
                    # -- save flag
                    with open(flag_fpath, "w") as fopen:
                        fopen.write(
                            f"FINISHED AT {time.asctime( time.localtime(time.time()) )}."
                        )
                    worker_status.append(True)
                else:
                    worker_status.append(False)
            else:
                with open(flag_fpath, "r") as fopen:
                    content = fopen.readlines()
                self._print(content)
                worker_status.append(True)

        output = workers
        if all(worker_status):
            if self.extract_data:
                self._print("--- extract results ---")
                status, computed_structures = extract_results_from_workers(
                    self.directory / "extracted",
                    workers,
                    safe_inspect=False,
                    use_archive=self.use_archive,
                    print_func=self._print,
                    debug_func=self._debug,
                )
                self.status = status
                if self.status:
                    self._print(f"extracted structures: {computed_structures}")
                    num_workers = len(workers)
                    reduce_single_worker = self.reduce_single_worker
                    if num_workers != 1:
                        reduce_single_worker = False
                    computed_structures = convert_results_to_structures(
                        computed_structures,
                        inp_shape,
                        inp_markers,
                        reduce_single_worker=reduce_single_worker,
                        merge_workers=self.merge_workers,
                        print_func=self._print,
                        debug_func=self._debug,
                    )
                    if self.check_scf_convergence:
                        selector = ScfSelector(scf_converged=True)
                        selector.directory = self.directory / "selected"
                        selector.select(
                            computed_structures,
                        )
                    output = computed_structures
                else:
                    ...
            else:
                self.status = "finished"
        else:
            ...

        return output


@registers.operation.register
class extract_cache(Operation):
    """Extract results from finished (cache) calculation wdirs.

    This is useful when reading results from manually created structures.

    """

    def __init__(
        self, compute, cache_wdirs: List[Union[str, pathlib.Path]], directory="./"
    ) -> None:
        """"""
        super().__init__(input_nodes=[compute], directory=directory)

        self.cache_wdirs = cache_wdirs

        return

    @CustomTimer(name="extract_cache", func=config._debug)
    def forward(self, workers: List[DriverBasedWorker]):
        """"""
        super().forward()

        # - broadcast workers
        nwdirs = len(self.cache_wdirs)
        nworkers = len(workers)
        assert (
            nwdirs == nworkers
        ) or nworkers == 1, "Found inconsistent number of cache dirs and workers."

        # - use driver to read results
        cache_data = self.directory / "cache_data.h5"
        if not cache_data.exists():
            from joblib import Parallel, delayed

            # TODO: whether check convergence?
            trajectories = Parallel(n_jobs=config.NJOBS)(
                delayed(self._read_trajectory)(curr_wdir, curr_worker)
                for curr_wdir, curr_worker in itertools.zip_longest(
                    self.cache_wdirs, workers, fillvalue=workers[0]
                )
            )
            trajectories = AtomsNDArray(data=trajectories)
            trajectories.save_file(cache_data)
        else:
            self._print("read cache...")
            trajectories = AtomsNDArray.from_file(cache_data)

        self.status = "finished"

        return trajectories

    @staticmethod
    def _read_trajectory(wdir, worker):
        """"""
        worker.driver.directory = wdir

        return worker.driver.read_trajectory()


@registers.operation.register
class extract(Operation):
    """Extract dynamics trajectories from a drive-node's worker."""

    def __init__(
        self,
        compute,
        merge_workers: bool = False,
        reduce_single_worker: bool = True,
        use_archive: bool = True,
        check_scf_convergence: bool = False,
        directory="./",
        *args,
        **kwargs,
    ) -> None:
        """Init an extract operation.

        Args:
            compute: Any node forwards a List of workers.
            merge_workers: Whether merge results from different workers togather.
            use_archive: Whether archive computation folders after all workers finished.

        """
        super().__init__(input_nodes=[compute], directory=directory)

        self.merge_workers = merge_workers
        self.reduce_single_worker = reduce_single_worker

        self.use_archive = use_archive
        self.check_scf_convergence = check_scf_convergence

        return

    def forward(self, workers: List[DriverBasedWorker]) -> AtomsNDArray:
        """
        Args:
            workers: ...

        Returns:
            AtomsNDArray.

        """
        super().forward()

        self.workers = workers  # for operations to access

        # - shape
        inp_shape, inp_markers = get_shape_data(
            self.input_nodes[0].directory / "_shape"
        )

        # - extract results
        status, computed_structures = extract_results_from_workers(
            self.directory,
            workers,
            use_archive=self.use_archive,
            print_func=self._print,
            debug_func=self._debug,
        )
        if status:
            self._print(f"extracted structures: {computed_structures}")
            num_workers = len(workers)
            reduce_single_worker = self.reduce_single_worker
            if num_workers != 1:
                reduce_single_worker = False
            computed_structures = convert_results_to_structures(
                computed_structures,
                inp_shape,
                inp_markers,
                reduce_single_worker=reduce_single_worker,
                merge_workers=self.merge_workers,
                print_func=self._print,
                debug_func=self._debug,
            )
            if self.check_scf_convergence:
                selector = ScfSelector(scf_converged=True)
                selector.directory = self.directory / "selected"
                selector.select(
                    computed_structures,
                )

        self.status = status

        return computed_structures


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import argparse

import numpy as np

import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
plt.style.use("presentation")

"""Command Line
        plumed sum_hills --stride 100 --hills ../hills.dat --mintozero
"""


def read_fesdat(fesDat):
    """"""
    fopen = open(fesDat, "r")

    # - parse fields...
    fields = fopen.readline().strip().split()
    ncolvar = int((len(fields) - 2) / 2.)

    # - read statistics
    cvDict = {}
    for idx in range(ncolvar):
        cvDict[idx] = {}
        dataLines = []
        for jdx in range(4):
            dataLines.append(fopen.readline())

        cvDict[idx]["min"] = float(dataLines[0].strip().split()[-1])
        cvDict[idx]["max"] = float(dataLines[1].strip().split()[-1])
        cvDict[idx]["nbins"] = int(dataLines[2].strip().split()[-1])
        cvDict[idx]["periodic"] = dataLines[3].strip().split()[-1]

    # - mark start point of the data
    dataLinePoint = fopen.tell()

    # get grids
    lastNumBins = 0
    for idx in range(ncolvar):
        fopen.seek(dataLinePoint)
        dataLists = []
        for jdx in range(cvDict[idx]['nbins']):
            dataLists.append(fopen.readline().strip().split())
            for kdx in range(lastNumBins):
                line = fopen.readline()
        dataList = [data[idx] for data in dataLists]
        cvDict[idx]['points'] = np.array(dataList, dtype=float)
        if lastNumBins == 0:
            lastNumBins = cvDict[idx]['nbins']
        else:
            lastNumBins = lastNumBins*cvDict[idx]['nbins']

    # free energy
    # TODOL: only 2 CV for now, maybe extended later
    freeEnergies = []
    if ncolvar == 1:
        fopen.seek(dataLinePoint)
        curList = []
        for idx in range(cvDict[0]['nbins']):
            curList.append(float(fopen.readline().strip().split()[1]))
        freeEnergies.extend(curList)
    elif ncolvar == 2:
        fopen.seek(dataLinePoint)
        for idx in range(cvDict[1]['nbins']):
            curList = []
            for jdx in range(cvDict[0]['nbins']):
                curList.append(float(fopen.readline().strip().split()[2]))
            freeEnergies.append(curList)
            emptyLine = fopen.readline()
    else:
        raise RuntimeError('cannot support number of CV larger than 2')

    freeEnergies = np.array(freeEnergies)

    fopen.close()

    return cvDict, freeEnergies

def plot_2D(datFile):
    """"""
    cvDict, free_energies = read_fesdat(datFile)
    free_energies *= kJMol2eV
    cv1, cv2 = cvDict[0]["points"], cvDict[1]["points"]
    #print(np.meshgrid(cv1, cv2))
    xv, yv = np.meshgrid(cv1, cv2)
    colvars = [(x, y) for x, y in zip(xv.flatten().tolist(), yv.flatten().tolist())]
    print(colvars)
    #print(f"Free Energies: {free_energies.shape}")
    #print(f"Free Energies: {free_energies}")
    #print(f"cv1: {cv1.shape}")
    #print(f"cv1: {cv1}")
    #print(f"cv2: {cv2.shape}")
    #print("cv2: {}".format(np.where(cv2 < 0., cv2+1., cv2)))
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8))
    ax.set_xlabel("CV 1")
    ax.set_ylabel("CV 2")

    xxx = np.zeros(free_energies.shape)
    xxx[:14, :] = free_energies[15:, :]
    xxx[14:, :] = free_energies[:15, :]
    cntr = ax.contourf(
        #cv1, cv2, free_energies, levels=10, 
        cv1, sorted(np.where(cv2 < 0., cv2+1., cv2).tolist()), xxx, levels=10, 
        #cmap="RdBu"
    )
    fig.colorbar(cntr, ax=ax, label="Free Energy [eV]")
    #cn = ax.contour(
    #    cv1, cv2, free_energies, #cmap="RdBu"
    #)
    #plt.clabel(cn, inline=True, fontsize=10)
    plt.savefig(os.path.splitext(os.path.basename(datFile))[0]+'.png')

    return


if __name__ == "__main__":
    """"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d", "--datafiles", nargs="*", default=["fes.dat"], 
        help="number of fes.dat files"
    )
    parser.add_argument(
        "--names", nargs="*", default=None, 
        help="names"
    )
    parser.add_argument(
        "-t", "--title", default="MetaDynamics Convergence Analysis", 
        help="title"
    )

    args = parser.parse_args()

    
    datList = args.datafiles
    ndatafiles = len(args.datafiles)
    if args.names is None:
        args.names = [str(i) for i in range(ndatafiles)]
    names = args.names

    kJMol2eV = 1/96.485
    nm2ang = 10.

    # check the CV dimensions
    cvDict, freeEnergies = read_fesdat(datList[0])

    # plot figure(s)
    if len(cvDict.keys()) == 1:
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8))
        ax.set_title(args.title)

        ax.set_xlabel("Collective Variable") 
        ax.set_ylabel("Free Energy [eV]") 
        for name, datFile in zip(names, datList):
            cvDict, freeEnergies = read_fesdat(datFile)
            ax.plot(cvDict[0]['points']*nm2ang, freeEnergies*kJMol2eV, label=name)
        ax.legend()
    elif len(cvDict.keys()) == 2:
        for datFile in datList:
            plot_2D(datFile)
    else:
        ...

    plt.savefig('fes.png')


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import itertools
import pathlib
from typing import NoReturn, List, Union

import numpy as np

from ase.io import read, write

from .. import config
from ..core.register import registers
from ..utils.command import parse_input_file


""""""


def run_newtrainer(configuration, directory):
    """"""
    config._print(f"{configuration = }")
    params = parse_input_file(configuration)

    # - create trainer
    name = params["trainer"].get("name", None)
    trainer = registers.create(
        "trainer", name, convert_name=True, **params["trainer"]
    )
    trainer.directory = directory

    # - create dataset
    name = params["dataset"].get("name", None)
    dataset = registers.create(
        "dataloader", name, convert_name=True, **params["dataset"]
    )

    # - other options
    init_model = params.get("init_model", None)

    # TODO: merge below two into one func?
    trainer.train(dataset, init_model=init_model)
    trainer.freeze()

    return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pickle

import numpy as np

from ase import Atoms
from ase import data, units

from .. import convert_string_to_atoms
from .move import MoveOperator
from .swap import SwapOperator
from .exchange import ExchangeOperator
from .react import ReactOperator

def save_operator(op, p):
    """"""
    with open(p, "wb") as fopen:
        pickle.dump(op, fopen)

    return

def load_operator(p):
    """"""
    with open(p, "rb") as fopen:
        op = pickle.load(fopen)

    return op

def select_operator(operators: list, probs: list, rng=np.random):
    """Select an operator based on the relative probabilities."""
    noperators = len(operators)
    op_idx = rng.choice(noperators, 1, probs)[0]
    op = operators[op_idx]

    return op

def parse_operators(op_params: dict):
    """Parse parameters for various operators.

    Currently, we have move, swap, and exchange (insert/remove).

    """
    operators, probs = [], []
    for param in op_params:
        name = param.pop("method", "move")
        prob = param.get("prob", 1.0)
        if name == "move":
            op = MoveOperator(**param)
        elif name == "swap":
            op = SwapOperator(**param)
        elif name == "exchange":
            op = ExchangeOperator(**param)
        elif name == "react":
            op = ReactOperator(**param)
        else:
            raise NotImplementedError(f"{name} is not supported.")
        operators.append(op)
        probs.append(prob)
    
    # - reweight probabilities
    probs = (np.array(probs) / np.sum(probs)).tolist()

    return operators, probs

def compute_thermo_wavelength(expart: str, temperature: float):
    # - beta
    kBT_eV = units.kB * temperature
    beta = 1./kBT_eV # 1/(kb*T), eV

    # - cubic thermo de broglie 
    hplanck = units._hplanck # J/Hz = kg*m2*s-1
    #_mass = np.sum([data.atomic_masses[data.atomic_numbers[e]] for e in expart]) # g/mol
    _species = convert_string_to_atoms(expart)
    _species_mass = np.sum(_species.get_masses())
    #print("species mass: ", _mass)
    _mass = _species_mass * units._amu
    kbT_J = kBT_eV * units._e # J = kg*m2*s-2
    cubic_wavelength = (hplanck/np.sqrt(2*np.pi*_mass*kbT_J)*1e10)**3 # thermal de broglie wavelength

    return _species_mass, beta, cubic_wavelength


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import collections
import itertools

import numpy as np

import jax
import jax.numpy as jnp

from ase.io import read, write
from ase.neighborlist import NeighborList, natural_cutoffs

from .gptools import compute_body3_descriptor


def switch_function(r, r_cut, gamma=2):
    """"""
    pr = r/r_cut

    return 1. - (1.+gamma)*pr**gamma + gamma*pr**(1+gamma)

def radial_function(r, r_span, width, r_cut, gamma):
    """Compute radial function.

    Args:
        r:
        r_span: shape (1, num_span)
        f_cut: shape (num_r, 1)

    """
    constant = 1./(r_cut/r_span.shape[1])
    f_cut = switch_function(r, r_cut, gamma)

    return constant*1./r**2*f_cut*jnp.exp(-(r-r_span)**2/2/width**2)

def compute_feature(positions, pair_indices, r_span, width, r_cut, gamma):
    """"""
    pair_positions = jnp.take(positions, pair_indices, axis=0)
    dis_vecs = pair_positions[0] - pair_positions[1]
    distances = jnp.linalg.norm(dis_vecs, axis=1)[:, jnp.newaxis] # (num_b2, 1)

    radial_values = radial_function(distances, r_span, width, r_cut, gamma)

    feature = jnp.sum(radial_values, axis=0)

    return feature

def benchmark(frames):
    """"""
    # - 
    r_cut = 6.0
    r_delta = 0.5

    # -
    (
        body2_mapping, body2_features, body2_gradients, 
        body3_mapping, body3_features, body3_gradients
    ) = compute_body3_descriptor(frames, r_cut=r_cut)

    positions = jnp.array(frames[0].positions)
    pair_indices = np.array([x[1:] for x in body2_mapping]).T
    print(pair_indices)

    r_span = jnp.linspace(-1., r_cut+2.0, num=13)[jnp.newaxis, :]

    feature = compute_feature(positions, pair_indices, r_span, width=r_delta, r_cut=r_cut, gamma=2)
    print(feature)

    feature_gradient = jax.jacfwd(compute_feature, argnums=0)(
        positions, pair_indices, r_span, width=r_delta, r_cut=r_cut, gamma=2
    )
    print(feature_gradient.shape)
    feature_drdx = jnp.sum(feature_gradient, axis=0)
    print(feature_drdx)
    print(feature_drdx.shape)

    #test_b2 = jnp.array([[1.28]])
    #print(switch_function(test_b2, r_cut, 2))
    #print(jax.jacfwd(switch_function, argnums=0)(test_b2, r_cut, 2))
    #xxx = jax.jacfwd(radial_function, argnums=0)(test_b2, r_span, r_delta, r_cut, 2)
    #print(xxx)

    return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np

from ..data.array import AtomsNDArray
from ..data.extatoms import ScfErrAtoms
from .selector import AbstractSelector


class ScfSelector(AbstractSelector):

    name = "scf"

    default_parameters = dict(scf_converged=True)

    def _mark_structures(self, data: AtomsNDArray, *args, **kwargs) -> None:
        """"""
        markers, structures = data.markers, data.get_marked_structures()
        if self.parameters["scf_converged"]:
            selected_indices = [
                i for i, a in enumerate(structures) if not isinstance(a, ScfErrAtoms)
            ]
        else:
            selected_indices = [
                i for i, a in enumerate(structures) if not isinstance(a, ScfErrAtoms)
            ]
        selected_markers = [markers[i] for i in selected_indices]

        data.markers = np.array(selected_markers)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
import pathlib
from typing import List

import numpy as np
from scipy.interpolate import make_interp_spline, BSpline

import matplotlib.pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    #print("Used default matplotlib style.")
    ...

from ase import Atoms
from ase.io import read, write
from ase.neighborlist import NeighborList

from .validator import AbstractValidator
from .utils import wrap_traj
from ..data.array import AtomsNDArray

def smooth_curve(bins, points):
    """"""
    spl = make_interp_spline(bins, points, k=3)
    bins = np.linspace(bins.min(), bins.max(), 300)
    points= spl(bins)

    for i, d in enumerate(points):
        if d < 1e-6:
            points[i] = 0.0

    return bins, points

def calc_rdf(
        wdir: pathlib.Path, frames, pairs, volume: float=None, 
        nbins: int=60, cutoff: float=6.0
    ) -> dict:
    """Calculate radial distribution.

    Args:
        wdir: Working directory that stores RDF results.
        frames: A List of Atoms objects.
        pairs: Target species pairs, for example, ["Cu-Cu", "Cu-O"].
        volume: System volume.
        nbins: Number of bins for histogram.
        cutoff: Cut-off radius in Angstrom.

    """
    if not wdir.exists():
        wdir.mkdir(parents=True)

    # --- parse system
    # NOTE: We assume the system volume does not change along the trajectory!
    if volume is None:
        volume = frames[0].get_volume()

    # NOTE: the atom order should be consistent in the entire trajectory
    #       i.e. this does not work for variable-composition system
    chemical_symbols = frames[0].get_chemical_symbols()
    sym_dict = {k: [] for k in set(chemical_symbols)}
    for k, v in itertools.groupby(enumerate(chemical_symbols), key=lambda x:x[1]):
        sym_dict[k].extend([x[0] for x in v])

    pair_dict = {}
    for pair in pairs:
        p0, p1 = pair.split("-")
        first_indices  = copy.deepcopy(sym_dict.get(p0, []))
        second_indices = copy.deepcopy(sym_dict.get(p1, []))
        assert len(first_indices) > 0, f"Cant found {p0}."
        assert len(second_indices) > 0, f"Cant found {p1}."
        #self._debug(f"first : {first_indices}")
        #self._debug(f"second: {second_indices}")

        if p0 == p1:
            avg_density = len(first_indices)/volume
        else:
            avg_density = (len(first_indices)+len(second_indices))/volume

        pair_dict[pair] = [first_indices, second_indices, avg_density]

    # -- What atoms we need access neighbour list
    combined_first_indices = set(itertools.chain(*[v[0] for k, v in pair_dict.items()]))

    # ---
    binwidth = cutoff/nbins
    bincentres = np.linspace(binwidth/2., cutoff+binwidth/2., nbins+1)
    left_edges = np.copy(bincentres) - binwidth/2.
    right_edges = np.copy(bincentres) + binwidth/2.
    bins = np.linspace(0., cutoff+binwidth, nbins+2)

    # ---
    dis_hist = {k: [] for k in pairs}
    for atoms in frames:
        cell = atoms.get_cell()
        symbols = atoms.get_chemical_symbols()
        positions = atoms.get_positions()
        natoms = len(atoms)
        nl = NeighborList(
            [(cutoff+binwidth)/2.]*natoms, skin=0., sorted=False,
            self_interaction=False, bothways=True,
        )
        nl.update(atoms)
        for x in combined_first_indices:
            nei_indices, nei_offsets = nl.get_neighbors(x)
            #distances, sel_indices = [], []
            distance_dict = {k: [] for k in pairs}
            for nei_ind, nei_off in zip(nei_indices, nei_offsets):
                #if nei_ind in second_indices:
                curr_pair = "-".join([symbols[s] for s in [x, nei_ind]])
                if curr_pair in pairs:
                    dis = np.linalg.norm(
                        positions[x] -
                        (positions[nei_ind] + np.dot(cell, nei_off))
                    )
                    distance_dict[curr_pair].append(dis)
            for k, v in distance_dict.items():
                hist_, edges_ = np.histogram(v, bins)
                dis_hist[k].append(hist_)

    # - reformat data
    results = {}
    for k, v in dis_hist.items():
        curr_dis_hist = np.array(v)
        avg_density = pair_dict[k][2]

        vshells = 4.*np.pi*left_edges**2*binwidth # NOTE: VMD likely uses this formula
        #vshells = 4./3.*np.pi*binwidth*(3*left_edges**2+3*left_edges*binwidth+binwidth**2)
        vshells[0] = 1. # avoid zero in division

        rdf = curr_dis_hist/vshells/avg_density

        rdf_avg = np.average(rdf, axis=0)
        rdf_min = np.min(rdf, axis=0)
        rdf_max = np.max(rdf, axis=0)
        rdf_svar = np.sqrt(np.var(rdf, axis=0))

        data = np.vstack((bincentres, rdf_avg, rdf_svar, rdf_min, rdf_max)).T
        np.savetxt(
            wdir/f"{k}.dat", data, 
            fmt="%8.4f  %8.4f  %8.4f  %8.4f  %8.4f", 
            header=("{:<8s}  "*5).format("r", "rdf", "svar", "min", "max")
        )
        results[k] = data

    return results


def plot_rdf(fig_path, data=None, ref_data=None, title="RDF"):
    """"""
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8))

    plt.suptitle("Radial Distribution Function")
    ax.set_xlabel("r [Å]")
    ax.set_ylabel("g(r)")

    ax.set_title(title)

    if data is not None:
        bincentres, rdf = data[:,0], data[:,1]
        bincentres_, rdf_ = smooth_curve(bincentres, rdf)
        ax.plot(bincentres_, rdf_, label="prediction")

    if ref_data is not None:
        bincentres, rdf = ref_data[:,0], ref_data[:,1]
        bincentres_, rdf_ = smooth_curve(bincentres, rdf)
        ax.plot(bincentres_, rdf_, ls="-.", label="reference")

    plt.legend()

    plt.savefig(fig_path)

    return


class RdfValidator(AbstractValidator):

    def __init__(self, pairs: List[str], cutoff: float=6., nbins: int=60, directory = "./", *args, **kwargs) -> None:
        """Radial Distribution.

        Args:
            paris: A List of species pairs [Cu-Cu, ..., ...].

        """
        super().__init__(directory=directory, *args, **kwargs)

        self.pairs = pairs
        #assert len(self.pair), f"{self.__class__.__name__} requires two elements."

        self.cutoff = cutoff
        self.nbins = nbins

        return

    def _process_data(self, data) -> List[List[Atoms]]:
        """"""
        data = AtomsNDArray(data)

        if data.ndim == 1:
            data = [data.tolist()]
        elif data.ndim == 2: # assume it is from extract_cache...
            data = data.tolist()
        elif data.ndim == 3: # assume it is from a compute node...
            data_ = []
            for d in data[:]: # TODO: add squeeze method?
                data_.extend(d)
            data = data_
        else:
            raise RuntimeError(f"Invalid shape {data.shape}.")

        return data[0] # TODO: support several trajectories

    def run(self, dataset, worker=None, *args, **kwargs):
        """Process reference and prediction data separately.

        TODO:

            Support average over several trajectories.

        """
        # - get custom volume
        volume = kwargs.get("volume", None)

        # - process dataset
        self._print("process reference ->")
        reference = dataset.get("reference", None)
        if reference is not None:
            ref_frames = self._process_data(reference)
            self._debug(f"reference  nframes: {len(ref_frames)}")
            ref_data = self._compute_rdf(
                self.directory/"reference", 
                ref_frames, self.pairs, self.cutoff, self.nbins,
                volume=volume
            )
        else:
            ref_data = None

        self._print("process prediction ->")
        prediction = dataset.get("prediction", None)
        if prediction is not None:
            pre_frames = self._process_data(prediction)
            self._debug(f"prediction nframes: {len(pre_frames)}")
            pre_data = self._compute_rdf(
                self.directory/"prediction", 
                pre_frames, self.pairs, self.cutoff, self.nbins,
                volume=volume
            )
        else:
            pre_data = None
        
        assert (ref_data is not None or pre_data is not None), "Neither reference nor prediction is given."

        # - compare results
        self._compare_results(ref_data, pre_data)

        return
    
    def _compute_rdf(self, wdir, frames: List[Atoms], pairs, cutoff, nbins, volume: float=None):
        """"""
        # - treat pbc
        if any(frames[0].pbc):
            frames = wrap_traj(frames)

        if not wdir.exists():
            data = calc_rdf(
                wdir, frames, pairs, volume, nbins, cutoff
            )
        else:
            data = {}
            saved_files = list(wdir.glob("*.dat"))
            for p in saved_files:
                data[p.name[:-4]] = np.loadtxt(p)

        return data
    
    def _compare_results(self, reference, prediction):
        """"""
        for pair in self.pairs:
            p = prediction.get(pair, None)
            r = reference.get(pair, None)
            if not (p is None and r is None):
                plot_rdf(
                    self.directory/f"{pair}_rdf.png", 
                    p, r, title=pair
                )

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import json
import logging
import pathlib
import time
import traceback

import yaml

import omegaconf
from omegaconf import OmegaConf

from .. import config
from .utils import create_variable, create_operation


class SessionInitialiser:

    #: Shared node objects.
    cache_nodes = {}

    @staticmethod
    def instantiate_operation(op_name, op_params):
        """"""
        params = {}
        for k, v in op_params.items():
            params[k] = v  # resolve one by one...
        op = create_operation(op_name, op_params)
    
        return op
    
    @staticmethod
    def resolve_operations(config: dict):
        """"""
        operations = {}
        for op_name, op_params in config.items():
            op = SessionInitialiser.instantiate_operation(op_name, op_params)
            operations[op_name] = op
            SessionInitialiser.cache_nodes[op_name] = op
    
        return operations
    
    @staticmethod
    def register_custom_resolvers():
        """"""
    
        # - add resolvers
        def create_vx_instance(vx_name, _root_):
            """"""
            if vx_name not in SessionInitialiser.cache_nodes:
                vx_params = OmegaConf.to_object(_root_.variables.get(vx_name))
                vx = create_variable(vx_name, vx_params)
                SessionInitialiser.cache_nodes[vx_name] = vx
                return vx
            else:
                return SessionInitialiser.cache_nodes[vx_name]
    
        OmegaConf.register_new_resolver("vx", create_vx_instance, use_cache=False)
    
        def create_op_instance(op_name: str, _root_):
            """"""
            if op_name not in SessionInitialiser.cache_nodes:
                op_params = OmegaConf.to_object(_root_.operations.get(op_name))
                op = create_operation(op_name, op_params)
                SessionInitialiser.cache_nodes[op_name] = op
                return op
            else:
                return SessionInitialiser.cache_nodes[op_name]
    
        OmegaConf.register_new_resolver("op", create_op_instance, use_cache=False)
    
        # --
        def read_json(input_file):
            with open(input_file, "r") as fopen:
                input_dict = json.load(fopen)
    
            return input_dict
    
        OmegaConf.register_new_resolver("json", read_json)
    
        def read_yaml(input_file):
            with open(input_file, "r") as fopen:
                input_dict = yaml.safe_load(fopen)
    
            return input_dict
    
        OmegaConf.register_new_resolver("yaml", read_yaml)
    
        return


def run_session_once(config_dict: dict, feed_command=None, directory="./"):
    """Configure session with omegaconfig."""
    # - set directory
    directory = pathlib.Path(directory)

    # - load configuration and resolve it
    conf = OmegaConf.create(config_dict)

    # - add placeholders and their directories
    if "placeholders" not in conf:
        conf.placeholders = {}
    if feed_command is not None:
        pairs = [x.split("=") for x in feed_command]
        for k, v in pairs:
            if v.isdigit():
                v = int(v)
            conf.placeholders[k] = v
    config._debug(f"YAML: {OmegaConf.to_yaml(conf)}")

    # - check operations and their directories
    for op_name, op_params in conf.operations.items():
        op_params["directory"] = str(directory / op_name)

    # - set variable directory
    if "variables" not in conf:
        conf.variables = {}
    for k, v_dict in conf.variables.items():
        v_dict["directory"] = str(directory / "variables" / k)
    # print("YAML: ", OmegaConf.to_yaml(conf))

    # - resolve sessions
    # container = OmegaConf.to_object(conf.sessions)
    # for k, v in container.items():
    #    print(k, v)

    try:
        operations = SessionInitialiser.resolve_operations(conf["operations"])
    except omegaconf.errors.InterpolationResolutionError as err:
        config._debug(traceback.format_exc())
        err_key = (str(err).strip().split("\n")[1]).strip().split(":")[1]
        config._print(f"FAILED TO PARSE `{err_key}` KEY.")
        err_info_tail = traceback.format_exc().split("\n")[-10:]
        for e in err_info_tail:
            config._print(f"{e}")
        exit()

    container = {}
    for k, v in conf["sessions"].items():
        container[k] = operations[v]

    # - run session
    names = conf.placeholders.get("names", None)
    if names is not None:
        session_names = [x.strip() for x in names.strip().split(",")]
    else:
        session_names = [None] * len(container)

    # - some imported packages change `logging.basicConfig`
    #   and accidently add a StreamHandler to logging.root
    #   so remove it...
    for h in logging.root.handlers:
        if isinstance(h, logging.StreamHandler) and not isinstance(
            h, logging.FileHandler
        ):
            logging.root.removeHandler(h)

    # - get session general configs
    sconfigs = conf.get("configs", {})

    exec_mode = sconfigs.get("mode", "basic")
    if exec_mode == "basic":  # sequential
        from .basic import Session

        session_states = [True]
        for i, (k, v) in enumerate(container.items()):
            n = session_names[i]
            if n is None:
                n = k
            entry_operation = v
            session = Session(directory=directory / n)
            session.run(entry_operation, feed_dict={})
    elif exec_mode == "active":
        from .active import ActiveSession

        assert len(container) == 1, "ActiveSession only accepts one operation."

        session_states = []
        for i, (k, v) in enumerate(container.items()):
            n = session_names[i]
            if n is None:
                n = k
            entry_operation = v
            session = ActiveSession(
                steps=sconfigs.get("steps", 2),
                reset_random_state=sconfigs.get("reset_random_state", False),
                reset_random_config=sconfigs.get("reset_random_config", ("init", 0)),
                directory=directory / n,
            )
            session.run(entry_operation, feed_dict={})
            # config._print(f"{session.state =}")
            session_states.append(session.is_finished())
            # config._print(f"{session.state =}")
    else:
        session_states = [False]
        raise RuntimeError(f"Unknown session type {exec_mode}.")

    return all(session_states)


def run_session(
    config_filepath: str,
    feed_command=None,
    timewait: float = -1.0,
    directory: str = "./",
):
    """Configure session with omegaconfig."""
    directory = pathlib.Path(directory)

    config_filepath = pathlib.Path(config_filepath)

    if config_filepath.suffix == ".json":
        with open(config_filepath, "r") as fopen:
            config_dict = json.load(fopen)
    elif config_filepath.suffix == ".yaml":
        with open(config_filepath, "r") as fopen:
            config_dict = yaml.safe_load(fopen)
    else:
        raise RuntimeError(f"Fail to load config `{str(config_filepath)}`")

    raw_config_dict = config_dict

    SessionInitialiser.register_custom_resolvers()

    # -
    if timewait > 0:
        for i in range(1000):
            SessionInitialiser.cache_nodes = {} # Clear cache before a new run.
            config._print(f"Monitor is running step {i}!!!")
            config_dict = copy.deepcopy(raw_config_dict)
            is_finished = run_session_once(config_dict, feed_command, directory)
            if is_finished:
                break
            else:
                config._print(f"Monitor is sleeping for {timewait} seconds.")
                time.sleep(timewait)
        else:
            config._print("Reach maximum monitor for-loop.")
    else:
        run_session_once(config_dict, feed_command, directory)

    return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import List, Callable

from .. import config
from ..operation import Operation


#: A List of valid session states.
SESSION_STATE_LIST: List[str] = [
    "StepToStart",
    "StepFinished",
    "StepToContinue",
    "StepBroken",
    "LoopToStart",
    "LoopFinished",
    "LoopConverged",
    "LoopUnConverged",
]

#: A List of finished session states.
FINISHED_SESSION_STATES: List[str] = [
    "StepBroken",
    "LoopFinished",
    "LoopConverged",
    "LoopUnConverged",
]


class AbstractSession:

    #: Session State.
    _state: str = "LoopToStart"

    #: Standard print function.
    _print: Callable = config._print

    #: Standard debug function.
    _debug: Callable = config._debug


    @property
    def state(self) -> str:
        """"""

        return self._state

    @state.setter
    def state(self, state: str):
        """"""
        assert state in SESSION_STATE_LIST, f"Invalid state `{state}` to assign."
        self._state = state

        return

    def is_finished(self) -> bool:
        """"""
        is_finished = False
        if self.state in FINISHED_SESSION_STATES:
            is_finished = True

        return is_finished

    def _process_operation(self, node: Operation):
        """"""
        if not node.is_about_to_exit():
            if node.is_ready_to_forward():  # All input nodes finished.
                node.inputs = [
                    input_node.output for input_node in node.input_nodes
                ]
                node.output = node.forward(*node.inputs)
            else:
                # - check whether this node' not ready due to previous nodes are broken.
                broken_states = []
                for input_node in node.input_nodes:
                    if isinstance(input_node, Operation):
                        broken_states.append(input_node.is_about_to_exit())
                    else:
                        broken_states.append(False)
                self._debug(f"{node.input_nodes =}")
                self._debug(f"{broken_states =}")
                is_broken = any(broken_states)
                if not is_broken:
                    self.state = "StepToContinue"
                    self._print("  wait previous nodes to finish...")
                else:
                    # The `broken` status is contagious
                    node.status = "exit"
                    self.state = "StepBroken"
                    self._print("  The current node is broken.")
        else:
            self.state = "StepBroken"
            self._print("  The current node is broken.")

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import pathlib
from typing import Union, List

import numpy as np
import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    ...

from ase import Atoms
from ase.constraints import FixAtoms, constrained_indices

from ..core.node import AbstractNode
from ..utils.comparision import get_properties, plot_parity, plot_distribution
from ..builder.constraints import parse_constraint_info


def set_constraint(atoms, cons_text):
    """"""
    atoms._del_constraints()
    mobile_indices, frozen_indices = parse_constraint_info(
        atoms, cons_text, ignore_ase_constraints=True, ret_text=False
    )
    if frozen_indices:
        atoms.set_constraint(FixAtoms(indices=frozen_indices))

    return atoms


class SinglePointComparator(AbstractNode):

    def __init__(self, directory: Union[str, pathlib.Path] = "./", random_seed: int = None, *args, **kwargs):
        super().__init__(directory, random_seed, *args, **kwargs)

        return

    def run(self, prediction, reference):
        """"""
        # - input shape should be (?, ?, ?)
        print(f"reference: {reference}")
        print(f"prediction: {prediction}")

        reference = reference.get_marked_structures()
        prediction = prediction.get_marked_structures()

        #for a in reference:
        #    print(constrained_indices(a, only_include=FixAtoms))
        #    break

        assert len(reference) == len(prediction), "Number of structures are inconsistent."

        nframes, rmse_ret = self._plot_comparison("spc", reference, prediction)
        self.write_data([["spc", nframes, rmse_ret]])

        return

    def write_data(self, data, fname: str="rmse.dat"):
        """"""
        # - check data file
        keys = ["ene", "frc"]
        for rmse_ret in [x[2] for x in data]:
            for k in rmse_ret.keys():
                if k not in keys:
                    keys.append(k)
        content_fmt = "{:<24s}  {:>8d}  " + "{:>8.4f}  {:>8.4f}  "*len(keys) + "\n"

        header_fmt = "{:<24s}  {:>8s}  " + "{:>8s}  {:>8s}  "*len(keys) + "\n"
        header_data = ["#prefix", "nframes"]
        for k in keys:
            header_data.extend([f"{k}_rmse", f"{k}_std"])
        header = header_fmt.format(*header_data)
        
        content = header
        for prefix, nframes, rmse_ret in data:
            cur_data = [prefix, nframes]
            for k in keys:
                v = rmse_ret.get(k, None)
                if v is None:
                    cur_data.extend([np.nan, np.nan])
                else:
                    cur_data.extend([v["rmse"], v["std"]])
            content += content_fmt.format(*cur_data)
        
        with open(self.directory/fname, "w") as fopen:
            fopen.write(content)
        self._print(content)

        return

    def _plot_comparison(self, prefix, ref_frames: List[Atoms], pred_frames: List[Atoms]):
        """"""
        if not (self.directory/prefix).exists():
            (self.directory/prefix).mkdir(parents=True)

        nframes = len(ref_frames)
        ref_symbols, ref_energies, ref_forces = get_properties(ref_frames, apply_constraint=True)
        ref_natoms = [len(a) for a in ref_frames]
        pred_symbols, pred_energies, pred_forces = get_properties(pred_frames, apply_constraint=True)
        
        # - figure
        fig, axarr = plt.subplots(
            nrows=1, ncols=2,
            gridspec_kw={"hspace": 0.3}, figsize=(16, 9)
        )
        axarr = axarr.flatten()
        plt.suptitle(f"{prefix} with nframes {nframes}")

        # -- energies
        ene_rmse = plot_parity(
            axarr[0], ref_energies, pred_energies, x_name="ene", weights=ref_natoms
        )

        # -- forces
        frc_rmse = plot_parity(
            axarr[1], ref_forces, pred_forces, x_name="frc", x_types=ref_symbols
        )

        #if (self.directory/f"{prefix}.png").exists():
        #    warnings.warn(f"Figure file {prefix} exists.", UserWarning)
        plt.savefig(self.directory/prefix/"rmse.png")
        plt.close()

        # plot distributions
        fig, axarr = plt.subplots(
            nrows=1, ncols=2,
            gridspec_kw={"hspace": 0.3}, figsize=(16, 9)
        )
        axarr = axarr.flatten()
        plt.suptitle(f"{prefix} with nframes {nframes}")

        plot_distribution(
            axarr[0], ref_energies, pred_energies, x_name="ene", weights=ref_natoms
        )
        plot_distribution(
            axarr[1], ref_forces, pred_forces, x_name="frc", x_types=ref_symbols
        )

        plt.savefig(self.directory/prefix/"dist.png")
        plt.close()

        # - save results to data file
        rmse_ret = {}
        x_rmse, x_rmse_names = ene_rmse
        for _rms, rms_name in zip(x_rmse, x_rmse_names):
            rmse_ret[rms_name] = _rms
        x_rmse, x_rmse_names = frc_rmse
        for _rms, rms_name in zip(x_rmse, x_rmse_names):
            rmse_ret[rms_name] = _rms

        return nframes, rmse_ret


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import List

from ase import Atoms

from .builder import StructureModifier
from .builder.group import create_a_group

""""""


class CleaveGroupModifier(StructureModifier):

    name: str = "cleave_group"

    def __init__(self, group, substrates=None, *args, **kwargs):
        """"""
        super().__init__(substrates, *args, **kwargs)

        self.group = group

        return
    
    def run(self, substrates=None, *args, **kwargs) -> List[Atoms]:
        """"""
        super().run(substrates=substrates, *args, **kwargs)

        frames = []
        for atoms in substrates:
            ainds = create_a_group(atoms, self.group)
            frames.append(atoms[ainds])

        return frames


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
from typing import List

from . import registers
from . import AbstractPotentialManager, DummyCalculator


"""This manager registers ALL bias calculators."""


class BiasManager(AbstractPotentialManager):

    name = "bias"
    implemented_backends = ("ase",)

    valid_combinations = (
        ("ase", "ase"),
        ("ase", "lammps"),
    )

    def __init__(self) -> None:
        """"""
        super().__init__()

        return

    def register_calculator(self, calc_params, *agrs, **kwargs) -> None:
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)

        # parse params
        calc_params = copy.deepcopy(calc_params)
        bias_method = calc_params.pop("method", None)
        assert bias_method is not None, "Bias must have a method."

        # check whether have a colvar key
        # FIXME: we need a better code structures to deal with colvar
        colvar_ = None
        for k, v in calc_params.items():
            if k == "colvar":
                cv_params = copy.deepcopy(v)
                cv_name = cv_params.pop("name")
                colvar_ = registers.create("colvar", cv_name, **cv_params)
                break
        else:
            ...
        if colvar_ is not None:
            calc_params["colvar"] = colvar_

        # instantiate calculator
        calc = DummyCalculator()
        if self.calc_backend == "ase":
            bias_cls = registers.bias[bias_method]
            if hasattr(bias_cls, "broadcast"):
                calc = bias_cls.broadcast(calc_params)
                num_calcs = len(calc)
                if num_calcs == 1:
                    calc = calc[0]
            else:
                calc = bias_cls(**calc_params)
        else:
            ...

        self.calc = calc

        return

    @staticmethod
    def get_bias_cls(backend, method):
        """"""

        return registers.bias[method]

    @staticmethod
    def broadcast(manager: "BiasManager") -> List["BiasManager"]:
        """"""
        calc_params = copy.deepcopy(manager.calc_params)
        calc_params["backend"] = manager.calc_backend
        # print(f"{calc_params =}")

        bias_cls = manager.get_bias_cls(calc_params["backend"], calc_params["method"])
        if hasattr(bias_cls, "broadcast_params"):
            broadcasted_params = bias_cls.broadcast_params(calc_params)
            # print(f"{broadcasted_params =}")
            managers = []
            for inp_dict in broadcasted_params:
                manager = BiasManager()
                manager.register_calculator(inp_dict)
                managers.append(manager)
        else:
            managers = [manager]

        return managers


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse

import numpy as np

from ase.io import read, write
from ase.geometry import find_mic


def write_neb_data(opt_images):
    """"""
    nimages = len(opt_images)
    
    energies = np.array([a.get_potential_energy() for a in opt_images])
    energies = energies - energies[0]
    
    differences = np.zeros(len(opt_images))
    init_pos = opt_images[0].get_positions()
    for i in range(1,nimages):
        a = opt_images[i]
        vector = a.get_positions() - init_pos
        vmin, vlen = find_mic(vector, a.get_cell())
        #differences[i] = np.linalg.norm(vlen)
        differences[i] = np.linalg.norm(vmin)
    
    # save results to file
    neb_data = "./neb.dat"
    content = ""
    for i in range(nimages):
        content += "{:4d}  {:10.6f}  {:10.6f}\n".format(
            i, differences[i], energies[i]
        )
    with open(neb_data, "w") as fopen:
        fopen.write(content)

    return


#dft_frames = read(
#    "/mnt/scratch2/users/40247882/validations/structures/surfaces-2x2/hcp2fcc_brg_mep.xyz",
#    ":"
#)
#write_neb_data(dft_frames)


def plot_comparasion(dft_file, mlp_file, comment=None):
    import matplotlib as mpl
    mpl.use('Agg') #silent mode
    from matplotlib import pyplot as plt
    plt.style.use('presentation')

    from scipy.interpolate import make_interp_spline

    fig, axarr = plt.subplots(
        nrows=1, ncols=1, 
        gridspec_kw={'hspace': 0.3}, figsize=(16,12)
    )
    plt.suptitle("Reaction Path Comparasion "+comment)

    axarr = [axarr]
    #axarr = axarr.flatten()

    #plt.suptitle("GCMC Evolution")
    ax = axarr[0]

    ax.set_xlabel("Reaction Coordinate [AA]")
    ax.set_ylabel("Potential Energy [eV]")

    # DFT data
    dft_data = np.loadtxt(dft_file, dtype=float)
    ax.scatter(dft_data[:,1], dft_data[:,2], marker="x", label="DFT")
    model = make_interp_spline(dft_data[:,1], dft_data[:,2], k=2)
    xs = np.linspace(dft_data[0,1], dft_data[-1,1], 500)
    ys = model(xs)
    ax.plot(xs, ys)

    # MLP data
    mlp_data = np.loadtxt(mlp_file, dtype=float)
    ax.scatter(mlp_data[:,1], mlp_data[:,2], marker="x", label="MLP")
    model = make_interp_spline(mlp_data[:,1], mlp_data[:,2], k=2)
    xs = np.linspace(mlp_data[0,1], mlp_data[-1,1], 500)
    ys = model(xs)
    ax.plot(xs, ys)
    if mlp_data.shape[1] == 4:
        apex = 2
        ax.errorbar(
            mlp_data[:,1], mlp_data[:,2], yerr=apex*mlp_data[:,3],
            elinewidth=4, linewidth=0,
            label="2*deviation"
        )

    ax.legend()

    plt.savefig("mep.png")

    return

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--dft"
    )
    parser.add_argument(
        "--mlp"
    )
    parser.add_argument(
        "-c", "--comment"
    )

    args = parser.parse_args()

    plot_comparasion(args.dft, args.mlp, args.comment)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import collections
import itertools

import numpy as np

from ase.io import read, write
from ase.neighborlist import NeighborList, natural_cutoffs


DimerData = collections.namedtuple("DimerData", ["fi", "i", "j", "pos_i", "pos_j", "shift"])

def compute_body3_descriptor(frames, r_cut: float):
    """"""
    # -- get neighbours
    dimer_list = []

    curr_atomic_index = 0
    for i_frame, atoms in enumerate(frames):
        #print(f"frame: {i_frame}")
        natoms = len(atoms)
        nl = NeighborList(
            cutoffs=[r_cut/2.]*natoms, skin=0.0, sorted=False,
            self_interaction=False, bothways=True
            #self_interaction=False, bothways=False # BUG?
        )
        nl.update(atoms)
        for i in range(natoms):
            nei_indices, nei_offsets = nl.get_neighbors(i)
            for j, offset in zip(nei_indices, nei_offsets):
                pos_i, pos_j = atoms.positions[i, :], atoms.positions[j, :]
                shift = np.dot(offset, atoms.get_cell()) # take negative
                #pos_vec = pos_i - pos_j + shift
                dimer = DimerData(
                    fi = i_frame,
                    i=curr_atomic_index+i, j=curr_atomic_index+j, 
                    pos_i=pos_i, pos_j=pos_j, shift=shift
                    #v=pos_vec, d=np.linalg.norm(pos_vec)
                )
                dimer_list.append(dimer)
        curr_atomic_index += natoms
    
    # - find 2-body
    body2_mapping = [[p.fi, p.i, p.j] for p in dimer_list]
    body2_vectors = np.array([p.pos_i-(p.pos_j+p.shift) for p in dimer_list])
    body2_features = np.linalg.norm(body2_vectors, axis=1)[:, np.newaxis] # (n_dimers, 1)
    body2_gradients = np.concatenate([body2_vectors, -body2_vectors], axis=-1)/body2_features
    body2_gradients = body2_gradients[:, np.newaxis, :] # (num_b2, 1, 6)
    #print("BODY-2: ")
    #print(body2_mapping)
    #print(body2_features)
    #print(body2_gradients)

    # - find 3-body
    trimer_list = []
    for k, v in itertools.groupby(dimer_list, key=lambda x: x.i):
        for pair0, pair1 in itertools.combinations(v, 2):
            pos_i, pos_j = pair0.pos_j + pair0.shift, pair1.pos_j + pair1.shift
            distance = np.linalg.norm(pos_i - pos_j)
            if 1e-8< distance <= r_cut: # j and k may be the same and distance is zero
                pair2 = DimerData(
                    fi=pair0.fi,
                    i=pair0.j, j=pair1.j,
                    pos_i=pos_i, pos_j=pos_j, shift=np.zeros(3)
                )
                trimer_list.append((pair0, pair1, pair2))
    body3_mapping, body3_vectors = [], []
    for pairs in trimer_list:
        body3_mapping.append([pairs[0].fi, pairs[0].i, pairs[0].j, pairs[1].j])
        body3_vectors.append([p.pos_i-(p.pos_j+p.shift) for p in pairs]) 
    if len(body3_vectors) > 0:
        body3_vectors = np.array(body3_vectors) # shape (n_trimers, 3, 3)
        body3_features = np.linalg.norm(body3_vectors, axis=2) # shape (n_trimers, 3)
        body3_gradients = np.concatenate([body3_vectors, -body3_vectors], axis=-1)/body3_features[:, :, np.newaxis]

        #for b3, f in zip(trimer_list, body3_features):
        #    print(b3[2].i, b3[2].j, b3[2].pos_i, b3[2].pos_j, b3[2].shift, f[2])
        #print(body3_features)
        #np.savetxt("b3.dat", body3_features, fmt="%8.2f")
    else:
        body3_features = []
        body3_gradients = []

    #print("BODY-3: ")
    #print(body3_mapping)
    #print(body3_vectors)
    #print(body3_features)
    #print("gradient: ", body3_gradients.shape)

    return body2_mapping, body2_features, body2_gradients, body3_mapping, body3_features, body3_gradients

if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*


import copy
import logging
import pathlib
import time

import numpy as np

from .. import config
from ..utils.command import parse_input_file
from ..core.variable import Variable
from ..core.operation import Operation
from ..core.register import registers
from ..worker.explore import ExpeditionBasedWorker
from ..scheduler.interface import SchedulerVariable


"""
"""

class ExpeditionVariable(Variable):

    def __init__(self, directory="./", **kwargs):
        """"""
        random_seed = kwargs.get("random_seed", None)
        if random_seed is None:
            random_seed = np.random.randint(0, 1e8)

        method = kwargs.pop("method", None)
        if "builder" in kwargs:
            builder = self._create_a_builder(kwargs["builder"], random_seed)
            kwargs["builder"] = builder

        expedition = registers.create(
            "expedition", method, convert_name=False, **kwargs
        )

        super().__init__(initial_value=expedition, directory=directory)

        return
    
    def _create_a_builder(self, builder: dict, random_seed: int):
        """"""
        # - builder
        if isinstance(builder, dict):
            builder_params = copy.deepcopy(builder)
            builder_method = builder_params.pop("method")
            builder = registers.create(
                "builder", builder_method, convert_name=False, **builder_params
            )
        else:  # variable
            builder = builder.value
            np.random.seed(random_seed)
        
        return builder


class explore(Operation):

    #: Whether to actively update some attrs.
    _active: bool = False

    def __init__(
        self, expedition, worker, scheduler=None, wait_time=60, 
        active: bool=False, directory="./", *args, **kwargs
    ) -> None:
        """"""
        if scheduler is None:
            scheduler = SchedulerVariable()
        input_nodes = [expedition, worker, scheduler]
        super().__init__(input_nodes, directory)

        self.wait_time = wait_time

        self._active = active

        return

    def forward(self, expedition, dyn_worker, scheduler):
        """Explore an expedition and forward results for further analysis.

        Returns:
            Workers that store structures.
        
        """
        super().forward()

        expeditions = [expedition]
        nexpeditions = len(expeditions)
        if self._active:
            curr_iter = int(self.directory.parent.name.split(".")[-1])
            if curr_iter > 0:
                self._print("    >>> Update seed_file...")
                for i in range(nexpeditions):
                    prev_wdir = (
                        self.directory.parent.parent / 
                        f"iter.{str(curr_iter-1).zfill(4)}" / 
                        self.directory.name
                    ) / f"expedition-{i}"
                    if hasattr(expedition, "update_active_params"):
                        expedition.update_active_params(prev_wdir)

        # -
        if hasattr(expedition, "register_worker"):
            expedition.register_worker(dyn_worker)
        
        # - run expedition with a worker
        worker = ExpeditionBasedWorker(expedition, scheduler)
        worker.directory = self.directory
        worker.wait_time = self.wait_time

        worker.run()
        worker.inspect(resubmit=True)

        basic_workers = []
        if worker.get_number_of_running_jobs() == 0:
            basic_workers = worker.retrieve(include_retrieved=True)
            self._debug(f"basic_workers: {basic_workers}")
            # for w in basic_workers:
            #     print(w.directory)
            self.status = "finished"
        else:
            ...

        return basic_workers


def run_expedition(exp_params: dict, wait: float=None, directory="./", potter=None):
    """"""
    directory = pathlib.Path(directory)

    if potter is not None:
        exp_params["worker"] = potter
    else:
        if "worker" not in exp_params:
            raise RuntimeError("Expedition must have a worker.")

    scheduler_params = exp_params.pop("scheduler", {})
    scheduler = SchedulerVariable(**scheduler_params).value

    #method = exp_params.pop("method")
    #expedition = registers.create("variable", method, convert_name=True, **exp_params).value
    expedition = ExpeditionVariable(directory=directory, **exp_params).value
    expedition.directory = directory
    if hasattr(expedition, "register_worker"):
        expedition.register_worker(exp_params["worker"])

    if scheduler.name == "local":
        if wait is not None:
            for i in range(1000):
                expedition.run()
                if expedition.read_convergence():
                    break
                time.sleep(wait)
                config._print(f"wait {wait} seconds...")
            else:
                ...
        else:
            expedition.run()
    else: # submit to queue
        worker = ExpeditionBasedWorker(
            expedition=expedition, scheduler=scheduler, directory=directory
        )
        worker.run()
        worker.inspect(resubmit=True)
        if worker.get_number_of_running_jobs() == 0:
            config._print("Expedition finished...")
        else:
            ...

    return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import List, Union

from ase import Atoms

from .builder import StructureModifier


class RepeatModifier(StructureModifier):

    name = "repeat"

    def __init__(self, repeat: Union[int,List[int]]=1, substrates=None, *args, **kwargs):
        """"""
        super().__init__(substrates, *args, **kwargs)

        self.repeat = repeat

        return
    
    def run(self, substrates: List[Atoms], size: int=1, *args, **kwargs):
        """"""
        super().run(substrates=substrates, *args, **kwargs)

        frames = []
        for substrate in self.substrates:
            frames.append(substrate.repeat(self.repeat))

        return frames


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools
from typing import Union, List

import numpy as np

from ase import Atoms

from ..data.array import AtomsNDArray
from .selector import AbstractSelector


class IntervalSelector(AbstractSelector):

    name = "interval"

    default_parameters = dict(
        period=1,
        include_first=True,
        include_last=False,
    )

    """This is a number-unaware selector.
    """

    def __init__(self, directory="./", *args, **kwargs) -> None:
        """"""
        super().__init__(directory, *args, **kwargs)

        return

    def _mark_structures(self, data: AtomsNDArray, *args, **kargs) -> None:
        """Select structures.

        Add unmasks to input trajectories.

        Args:
            inp_dat: Structures.

        """
        # if axis == -1 or axis == ndim-1:
        #    # NOTE: Last dimension is the trajectory
        #    #       it may have padded dummy atoms
        #    ...
        # else:
        #    ...

        # - group markers
        if self.axis is not None:
            axis = self.axis
            ndim = len(data.shape)
            if axis < -ndim or axis > ndim:
                raise IndexError(f"axis {axis} is out of dimension {ndim}.")
            if axis < 0:
                axis = ndim + axis

            marker_groups = {}
            for k, v in itertools.groupby(data.markers, key=lambda x: x[axis]):
                if k in marker_groups:
                    marker_groups[k].extend(list(v))
                else:
                    marker_groups[k] = list(v)
        else:
            marker_groups = dict(all=data.markers)

        self._debug(f"marker_groups: {marker_groups}")

        selected_markers = []
        for curr_grpname, curr_markers in marker_groups.items():
            curr_markers = sorted(np.array(curr_markers).tolist())
            nstructures = len(curr_markers)

            first, last = 0, nstructures - 1
            if self.include_first:
                curr_indices = list(range(0, nstructures, self.period))
                if self.include_last:
                    if last not in curr_indices:
                        curr_indices.append(last)
                else:
                    if last in curr_indices:
                        curr_indices.remove(last)
            else:
                curr_indices = list(range(1, nstructures, self.period))
                if self.include_last:
                    if last not in curr_indices:
                        curr_indices.append(last)
                else:
                    if last in curr_indices:
                        curr_indices.remove(last)
            curr_selected_markers = [curr_markers[i] for i in curr_indices]
            selected_markers.extend(curr_selected_markers)

        data.markers = np.array(selected_markers)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
from typing import Optional, List

import numpy as np

from ase import Atoms
from ase.calculators.calculator import Calculator, all_changes
from ase.geometry import find_mic


def compute_harmonic_energy_and_forces(
    cell, positions, harmonic_position, kspring: float, selected_indices: List[int],
    pbc: bool=True
):
    """"""
    # - compute forces
    distances, vectors = [], []
    for i in selected_indices:
        pos_i = copy.deepcopy(positions[i])
        pos_i[2] = 0.
        vec, dis = find_mic(harmonic_position - pos_i, cell, pbc=pbc)
        vectors.append(vec)
        distances.append(dis)
    distances = np.array(distances)

    # - compute energy
    energy = np.sum(0.5*kspring*distances**2)

    # - compute forces
    forces = np.zeros(positions.shape)
    for i, vec in zip(selected_indices, vectors):
        frc_i = kspring*vec
        forces[i] += frc_i

    return energy, forces


class PlaneHarmonicCalculator(Calculator):

    implemented_properties = ["energy", "free_energy", "forces"]

    def __init__(self, harmonic_position, kspring: float = 0.1, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

        #: Spring constant, eV/Ang^2.
        self.kspring = kspring

        #: Harmonic position on the plane.
        self.harmonic_position = np.array(harmonic_position)
        assert self.harmonic_position[2] == 0.

        return

    def calculate(
        self,
        atoms: Optional[Atoms] = None,
        properties=["energy"],
        system_changes=["positions"],
    ):
        """"""
        super().calculate(atoms, properties, system_changes)

        target_indices = [45, 46, 47]

        energy, forces = compute_harmonic_energy_and_forces(
            atoms.cell, atoms.positions, self.harmonic_position, self.kspring, 
            selected_indices=target_indices, pbc=True
        )

        print(f"{energy =}")
        print(f"{forces[target_indices, :] =}")

        self.results["energy"] = energy
        self.results["free_energy"] = energy
        self.results["forces"] = forces

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*


"""Create scheduler based on parameters

This module includes several schedulers.

Example:

    .. code-block:: python

        >>> from gdpx.scheduler.local import LocalScheduler
        >>> params = dict()
        >>> scheduler = LocalScheduler(**params)

"""

from ..core.register import registers

from .local import LocalScheduler

registers.scheduler.register(LocalScheduler)

from .lsf import LsfScheduler

registers.scheduler.register(LsfScheduler)

from .pbs import PbsScheduler

registers.scheduler.register(PbsScheduler)

from .slurm import SlurmScheduler

registers.scheduler.register(SlurmScheduler)


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
from typing import NoReturn, List

import numpy as np
import networkx as nx
from joblib import Parallel, delayed

from ase import Atoms
from ase.io import read, write

from .. import config
from .. import CustomTimer
from .. import StruGraphCreator, extract_chem_envs
from .. import paragroup_unique_chem_envs

from ..builder import StructureModifier
from ..group import create_a_group

DEFAULT_GRAPH_PARAMS = dict(
    pbc_grid = [2, 2, 0],
    graph_radius = 2,
    neigh_params = dict(
        covalent_ratio = 1.1,
        skin = 0.25
    )
)

def single_create_structure_graph(graph_params: dict, target_group: List[str], atoms: Atoms) -> List[nx.Graph]:
    """Create structure graph and get selected chemical environments.

    Find atoms with selected chemical symbols or in the defined region.

    Args:
        graph_params: Parameters for the graph representation.
        target_indices: A List of Integers.
        atoms: Input structure.

    Returns:
        A list of graphs that represent the chemical environments of selected atoms.

    """
    stru_creator = StruGraphCreator(**graph_params)

    natoms = len(atoms)
    group_indices = list(range(natoms))
    for command in target_group:
        curr_indices = create_a_group(atoms, command)
        group_indices = [i for i in group_indices if i in curr_indices]
    #config._print(f"{group_indices = }")

    graph = stru_creator.generate_graph(atoms, ads_indices=group_indices)

    graph_radius = stru_creator.graph_radius
    chem_envs = extract_chem_envs(graph, atoms, group_indices, graph_radius)
    #config._print(f"{graph_radius = }")

    return chem_envs


class GraphModifier(StructureModifier):

    
    def run(self, substrates: List[Atoms]=None, size: int=1, *args, **kwargs):
        """"""
        super().run(substrates=substrates, *args, **kwargs)

        prev_directory = self.directory
        curr_substrates = self.substrates

        for i in range(size):
            self.directory = prev_directory/f"graph-{i}"
            self.directory.mkdir(parents=True, exist_ok=True)
            cached_filepath = self.directory/"enumerated.xyz"
            if not cached_filepath.exists():
                self._print("-- run graph results --")
                modified_structures = self._irun(curr_substrates, )
                write(self.directory/"enumerated.xyz", modified_structures)
            else:
                self._print("-- use cached results --")
                modified_structures = read(self.directory/"enumerated.xyz", ":")
            n_structures = len(modified_structures)
            self._print(f"nframes: {n_structures}")
            curr_substrates = modified_structures
        
        self.directory = prev_directory

        return modified_structures 
    
    def _irun(substrates: List[Atoms]) -> List[Atoms]:
        """"""

        raise NotImplementedError()

    def _compare_structures(self, ret_frames: List[Atoms], graph_params, spec_params):
        """"""
        with CustomTimer(name="create-graphs", func=self._print):
            ret = Parallel(n_jobs=self.njobs)(
                delayed(single_create_structure_graph)(graph_params, spec_params, a)
                for a in ret_frames
            )
        # not unique across substrates
        #write(self.directory/f"possible_frames-{self.op_num}.xyz", ret_frames)

        # - check if the ret is empty
        #   it happens when all species are removed/exchanged...
        ret_envs = []
        for x in ret:
            ret_envs.extend(x)

        if ret_envs:
            ret_env_groups = ret
            self._print("Typical Chemical Environment "+str(ret_envs[0]))
            with CustomTimer(name="check-uniqueness", func=self._print):
                # compare chem envs
                #unique_envs, unique_groups = unique_chem_envs(
                #    chem_groups, list(enumerate(frames))
                #)
                unique_envs, unique_groups = paragroup_unique_chem_envs(
                    ret_env_groups, list(enumerate(ret_frames)), directory=self.directory, 
                    #n_jobs=self.njobs
                    n_jobs=1
                )
                #self._print("number of unique groups: ", len(unique_groups))

            # - get unique structures
            created_frames = [] # graphly unique
            for x in unique_groups:
                created_frames.append(x[0][1])
            ncandidates = len(created_frames)

            # -- unique info
            unique_data = []
            for i, x in enumerate(unique_groups):
                data = ["ug"+str(i)]
                data.extend([a[0] for a in x])
                unique_data.append(data)
            content = "# unique, indices\n"
            content += f"# ncandidates {ncandidates}\n"
            for d in unique_data:
                content += ("{:<8s}  "+"{:<8d}  "*(len(d)-1)+"\n").format(*d)

            unique_info_path = self.directory / f"unique-info.txt"
            with open(unique_info_path, "w") as fopen:
                fopen.write(content)
        else:
            self._print("Cant find valid species...")
            created_frames = ret_frames
            ncandidates = len(created_frames)

        return created_frames
    

if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import copy
import dataclasses
import pathlib
import re
import shutil
import tarfile
import warnings

from typing import Optional, NoReturn, List, Callable, Union
from collections.abc import Iterable

import numpy as np

from ase import Atoms
from ase.constraints import FixAtoms
from ase.calculators.calculator import compare_atoms

from ..builder.constraints import parse_constraint_info, convert_indices
from ..core.node import AbstractNode

#: Prefix of backup files
BACKUP_PREFIX_FORMAT: str = "gbak.{:d}."

#: Parameter keys used to init a minimisation task.
MIN_INIT_KEYS: List[str] = ["min_style", "min_modify", "dump_period"]

#: Parameter keys used to run a minimisation task.
MIN_RUN_KEYS: List[str] = ["steps", "fmax"]

#: Parameter keys used to init a molecular-dynamics task.
MD_INIT_KEYS: List[str] = [
    "md_style", "velocity_seed", "timestep", "temp", "Tdamp", 
    "press", "Pdamp", "dump_period"
]


@dataclasses.dataclass
class Controller:

    #: Thermostat name.
    name: str = "controller" # thermostat or barostat

    #: Parameters.
    params: dict = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class DriverSetting:

    """These are geometric parameters. Electronic?
    """

    #: Simulation task.
    task: str = "min"

    #: Driver setting.
    backend: str = "external"

    #: 
    min_style: str = "bfgs"
    min_modify: str = "integrator verlet tmax 4"
    maxstep: float = 0.1

    #:
    md_style: str = "nvt"

    velocity_seed: int = None

    #: Whether ignore atoms' velocities and initialise it from the scratch.
    ignore_atoms_velocities: bool = False

    #: Whether remove rotation when init velocity.
    remove_rotation: bool = True

    #: Whether remove translation when init velocity.
    remove_translation: bool = True

    timestep: float = 1.0

    temp: float = 300.
    tend: float = None
    Tdamp: float = 100. # fs

    press: float = 1.0 # bar
    pend: float = None # bar
    Pdamp: float = 100.

    #: The interval steps to dump output files (e.g. trajectory).
    dump_period: int = 1

    #: The interval steps to save a check point that is used for restart.
    ckpt_period: int = 100

    #: The number of checkpoints to save.
    ckpt_number: int = 3

    #: run params
    etol: float = None # 5e-2
    fmax: float = None # 1e-5
    steps: int = 0

    constraint: str = None

    #: Parameters that are used to update 
    _internals: dict = dataclasses.field(default_factory=dict)

    def update(self, **kwargs):
        """"""
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
        self.__post_init__()

        return
    
    def get_init_params(self):
        """"""

        return copy.deepcopy(self._internals)
    
    def get_run_params(self, *args, **kwargs):
        """"""
        raise NotImplementedError(f"{self.__class__.__name__} has no function for run params.")

class AbstractDriver(AbstractNode):

    #: Driver's name.
    name: str = "abstract"

    #: Atoms that is for state check.
    atoms: Optional[Atoms] = None

    #: Whether check the dynamics is converged, and re-run if not.
    ignore_convergence: bool = False

    #: Whether accepct the bad structure due to crashed FF or SCF-unconverged DFT.
    accept_bad_structure: bool = False

    #: Driver setting.
    setting: DriverSetting = None

    #: List of output files would be saved when restart.
    saved_fnames: List[str] = []

    #: List of output files would be removed when restart.
    removed_fnames: List[str] = []

    #: Systemwise parameter keys.
    syswise_keys: list = []

    #: Parameters for PotentialManager.
    pot_params: dict = None

    def __init__(
        self, calc, params: dict, directory="./", ignore_convergence: bool=False, 
        random_seed=None, *args, **kwargs
    ):
        """Init a driver.

        Args:
            calc: The ase calculator.
            params: Driver parameters.
            directory: Working directory.

        """
        super().__init__(
            directory=directory, random_seed=random_seed, *args, **kwargs
        )

        self.calc = calc
        self.calc.reset()

        self.ignore_convergence = ignore_convergence

        self._org_params = copy.deepcopy(params)

        return
    
    @property
    @abc.abstractmethod
    def default_task(self) -> str:
        """Default simulation task."""

        return

    @property
    @abc.abstractmethod
    def supported_tasks(self) -> List[str]:
        """Supported simulation tasks"""

        return
    
    @AbstractNode.directory.setter
    def directory(self, directory_):
        """"""
        self._directory = pathlib.Path(directory_)
        # NOTE: directory is set before self.calc is defined...
        #       ASE uses str path, so to avoid inconsistency here
        if hasattr(self, "calc"):
            self.calc.directory = str(self.directory) 

        return
    
    def get(self, key):
        """Get param value from init/run params by a mapped key name."""
        parameters = copy.deepcopy(self.init_params)
        parameters.update(copy.deepcopy(self.run_params))

        value = parameters.get(key, None)
        if not value:
            mapped_key = self.param_mapping.get(key, None)
            if mapped_key:
                value = parameters.get(mapped_key, None)

        return value
    
    def reset(self) -> None:
        """Remove results stored in dynamics calculator."""
        self.calc.reset()

        return

    def run(self, atoms, read_ckpt: bool=True, extra_info: dict=None, *args, **kwargs) -> Atoms:
        """Return the last frame of the simulation.

        Copy input atoms, and return a new atoms. Check whether the simulation is
        finished and retrieve stored results. If necessary, extra information could 
        be added to the atoms.info.

        The simulation should either run from the scratch or restart from a given 
        checkpoint...

        """
        # - NOTE: input atoms from WORKER may have minimal properties as
        #         cell, pbc, positions, symbols, tags, momenta...
        atoms = atoms.copy()

        # - set driver's atoms to the current one
        if isinstance(self.atoms, Atoms):
            warnings.warn("Driver has attached atoms object.", RuntimeWarning)
            system_changes = compare_atoms(atoms1=self.atoms, atoms2=atoms, tol=1e-15)
            self._debug(f"system_changes: {system_changes}")
            self._debug(f"atoms to compare: {self.atoms} {atoms}")
            if len(system_changes) > 0:
                system_changed = True
            else:
                system_changed = False
        else:
            system_changed = False

        # - backup old params
        params_old = copy.deepcopy(self.calc.parameters)

        # - run dynamics
        cache_traj = None
        if not self._verify_checkpoint():
            # If there is no valid checkpoint, just run the simulation from the scratch
            self._debug(f"... start from the scratch @ {self.directory.name} ...")
            self.directory.mkdir(parents=True, exist_ok=True)
            self._irun(atoms, *args, **kwargs)
        else:
            # If there is valid checkpoints...
            if not system_changed:
                self._debug(f"... system not changed @ {self.directory.name} ...")
                cache_traj = self.read_trajectory()
                converged = self.read_convergence(cache_traj=cache_traj)
                self._debug(f"... convergence {converged} ...")
                if not converged:
                    self._debug(f"... unconverged @ {self.directory.name} ...")
                    ckpt_wdir = self._save_checkpoint() if read_ckpt else None
                    self._debug(f"... checkpoint @ {str(ckpt_wdir)} ...")
                    self._cleanup()
                    self._irun(atoms, ckpt_wdir=ckpt_wdir, cache_traj=cache_traj, *args, **kwargs)
                    cache_traj = None
                else:
                    self._debug(f"... converged @ {self.directory.name} ...")
            else:
                self._debug(f"... clean up @ {self.directory.name} ...")
                self._cleanup()
                self._irun(atoms, *args, **kwargs)

        self.calc.parameters = params_old
        self.calc.reset()
        
        # - check again
        cache_traj = self.read_trajectory()
        curr_atoms, converged = None, self.read_convergence(cache_traj=cache_traj) # NOTE: This will read_trajectory
        if converged:
            self._debug(f"... 2. converged @ {self.directory.name} ...")
            traj = cache_traj
            nframes = len(traj)
            if nframes > 0:
                curr_atoms = traj[-1]
                if extra_info is not None:
                    curr_atoms.info.update(**extra_info)
            else:
                warnings.warn(f"The calculation at {self.directory.name} performed but failed.", RuntimeWarning)
                curr_atoms = None
        else:
            self._debug(f"... 2. unconverged @ {self.directory.name} ...")

        return curr_atoms # TODO: change None to ...
    
    def _verify_checkpoint(self, *args, **kwargs) -> bool:
        """Check whether there is a previous calculation in the `self.directory`."""

        return self.directory.exists()
    
    def _save_checkpoint(self, *args, **kwargs):
        """Save the previous simulation to a checkpoint directory."""
        # - find previous runs...
        prev_wdirs = sorted(self.directory.glob(r"[0-9][0-9][0-9][0-9][.]run"))
        self._debug(f"prev_wdirs: {prev_wdirs}")
        curr_index = len(prev_wdirs)

        curr_wdir = self.directory/f"{str(curr_index).zfill(4)}.run"
        self._debug(f"curr_wdir: {curr_wdir}")

        # - backup files
        curr_wdir.mkdir()
        for x in self.directory.iterdir():
            if not re.match(r"[0-9]{4}\.run", x.name):
                #if x.name in self.saved_fnames:
                #    shutil.move(x, curr_wdir)
                #else:
                #    x.unlink()
                shutil.move(x, curr_wdir) # save everything...
            else:
                ...

        return curr_wdir
    
    @abc.abstractmethod
    def _irun(self, atoms: Atoms, *args, **kwargs):
        """Prepare input structure (atoms) and parameters and run the simulation."""

        return
    
    def _cleanup(self):
        """Remove unnecessary files.

        Some dynamics will not overwrite old files so cleanup is needed.

        """
        # retain calculator-related files
        for fname in self.removed_fnames:
            curr_fpath = self.directory/fname
            if curr_fpath.exists():
                curr_fpath.unlink()

        return

    def _preprocess_constraints(self, atoms: Atoms, run_params: dict) -> None:
        """Remove existing constraints on atoms and add FixAtoms.

        If have cons in kwargs overwrite current cons stored in atoms.

        """
        # - check constraint
        cons_text = run_params.pop("constraint", None)
        if cons_text is not None: # FIXME: check cons_text in parse_?
            atoms._del_constraints()
            mobile_indices, frozen_indices = parse_constraint_info(
                atoms, cons_text, ignore_ase_constraints=True, ret_text=False
            )
            if frozen_indices:
                atoms.set_constraint(FixAtoms(indices=frozen_indices))
            else:
                ...
        else:
            ...

        return
    
    def read_convergence(self, cache_traj: List[Atoms]=None, *args, **kwargs) -> bool:
        """Read output to check whether the simulation is converged.

        TODO:
            If not converged, specific params in input files should be updated.

        """
        if self.ignore_convergence:
            return True

        # - check whether the driver is coverged
        if cache_traj is None:
            traj_frames = self.read_trajectory() # NOTE: DEAL WITH EMPTY FILE ERROR
        else:
            traj_frames = cache_traj

        # - check if this structure is bad
        is_badstru = False
        for a in traj_frames:
            curr_is_badstru = a.info.get("is_badstru", False)
            if curr_is_badstru:
                is_badstru = True
                break
        else:
            ...

        if self.accept_bad_structure:
            return True
        else:
            if is_badstru:
                return False
            else:
                ...

        # - check actual convergence
        nframes = len(traj_frames)

        converged = False
        if nframes > 0:
            if self.setting.steps > 0:
                step = traj_frames[-1].info["step"]
                self._debug(f"nframes: {nframes}")
                if self.setting.task == "min":
                    # NOTE: check geometric convergence (forces)...
                    #       some drivers does not store constraints in trajectories
                    # NOTE: Sometimes constraint changes if `lowest` is used.
                    frozen_indices = None
                    run_params = self.setting.get_run_params()
                    cons_text = run_params.pop("constraint", None)
                    mobile_indices, beg_frozen_indices = parse_constraint_info(traj_frames[0], cons_text, ret_text=False)
                    if beg_frozen_indices:
                        frozen_indices = beg_frozen_indices
                    end_atoms = traj_frames[-1]
                    if frozen_indices:
                        mobile_indices, end_frozen_indices = parse_constraint_info(end_atoms, cons_text, ret_text=False)
                        if convert_indices(end_frozen_indices) != convert_indices(beg_frozen_indices):
                            self._print("Constraint changes after calculation, which may be from `lowest`. Most times it is fine.")
                        end_atoms._del_constraints()
                        end_atoms.set_constraint(FixAtoms(indices=frozen_indices))
                    # TODO: Different codes have different definition for the max force
                    maxfrc = np.max(np.fabs(end_atoms.get_forces(apply_constraint=True)))
                    if maxfrc <= self.setting.fmax or step+1 >= self.setting.steps:
                        converged = True
                    self._debug(
                        f"MIN convergence: {converged} STEP: {step+1} >=? {self.setting.steps} MAXFRC: {maxfrc} <=? {self.setting.fmax}"
                    )
                elif self.setting.task == "md":
                    if step+1 >= self.setting.steps: # step startswith 0
                        converged = True
                    self._debug(f"MD convergence: {converged} STEP: {step+1} >=? {self.setting.steps}")
                else:
                    raise NotImplementedError("Unknown task in read_convergence.")
            else:
                # just spc, only need to check force convergence
                if nframes == 1:
                    converged = True
            # self._print(f"energy: {traj_frames[-1].get_potential_energy():12.4f}")
        else:
            ...

        return converged

    @abc.abstractmethod
    def read_trajectory(self, *args, **kwargs) -> List[Atoms]:
        """Read trajectory in the current working directory.
        """

        return
    
    def _aggregate_trajectories(self, archive_path=None) -> List[Atoms]:
        """"""
        prev_wdirs = []
        if archive_path is None:
            prev_wdirs = sorted(self.directory.glob(r"[0-9][0-9][0-9][0-9][.]run"))
        else:
            pattern = self.directory.name + "/" + r"[0-9][0-9][0-9][0-9][.]run"
            with tarfile.open(archive_path, "r:gz") as tar:
                for tarinfo in tar:
                    if tarinfo.isdir() and re.match(pattern, tarinfo.name):
                        prev_wdirs.append(tarinfo.name)
            prev_wdirs = [self.directory/pathlib.Path(p).name for p in sorted(prev_wdirs)]
        self._debug(f"prev_wdirs: {prev_wdirs}")

        all_wdirs = prev_wdirs + [self.directory]

        traj_list = []
        for w in all_wdirs:
            curr_frames = self._read_a_single_trajectory(w, archive_path=archive_path)
            if curr_frames:
                traj_list.append(curr_frames)

        # -- concatenate
        # NOTE: For DFT calculations,
        #       some spin systems may give different scf convergence on the same
        #       structure. Sometimes, the preivous failed but the next run converged,
        #       The concat below uses the previous one...
        traj_frames, ntrajs = [], len(traj_list)
        if ntrajs > 0:
            traj_frames.extend(traj_list[0])
            for i in range(1, ntrajs):
                assert np.allclose(
                    traj_list[i - 1][-1].positions, traj_list[i][0].positions
                ), f"Traj {i-1} and traj {i} are not consecutive."
                traj_frames.extend(traj_list[i][1:])
        else:
            ...

        return traj_frames
    
    def as_dict(self) -> dict:
        """Return parameters of this driver."""
        params = dict(
            backend = self.name,
            ignore_convergence = self.ignore_convergence,
            random_seed=self.random_seed
        )
        # NOTE: we use original params otherwise internal param names would be 
        #       written out and make things confusing
        #       org_params are merged params thatv have init and run sections
        org_params = copy.deepcopy(self._org_params)

        # - update some special parameters
        constraint = self.setting.constraint
        org_params["constraint"] = constraint

        params.update(org_params)

        return params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
import pathlib
from typing import Union, List
import warnings

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.calculators.singlepoint import SinglePointCalculator

from ..core.register import registers
from ..core.variable import Variable
from ..core.node import AbstractNode

from .utils import is_a_valid_system_name


#: How to map keys in structures.
DEFAULT_PROP_MAP_KEYS: List[List[str]] = [("energy", "energy"), ("forces", "forces")]


def map_atoms_data(atoms: Atoms, prop_map_keys) -> None:
    """"""
    assert type(atoms.calc) == SinglePointCalculator, f"Atoms {atoms} has {atoms.calc}."

    # For some small dataset, structures can be used both as train and test,
    # thus, we need add a flag to avoid repeated mapping.
    is_mapped = atoms.info.get("is_mapped", False)

    if not is_mapped:
        results = copy.deepcopy(atoms.calc.results)
        for mapping_pairs in prop_map_keys:
            dst_key, src_key = mapping_pairs
            if src_key in results:
                # NOTE: The dst_data is overwritten if it exists
                src_val = results.pop(src_key)
                results[dst_key] = src_val
            else:
                raise KeyError(f"Atoms {atoms} has no key {src_key}.")
        atoms.calc.results = results
        atoms.info["is_mapped"] = True
    else:
        ...

    return


def traverse_xyzdirs(wdir):
    """"""
    data_dirs = []

    def recursive_traverse(wdir):
        for p in wdir.iterdir():
            if p.is_dir():
                xyzpaths = list(p.glob("*.xyz"))
                if len(xyzpaths) > 0:
                    data_dirs.append(p)
                recursive_traverse(p)
            else:
                ...
        return

    recursive_traverse(wdir)

    return data_dirs


class AbstractDataloader(AbstractNode): ...


@registers.dataloader.register
class XyzDataloader(AbstractDataloader):

    name = "xyz"

    """A directory-based dataset.

    There are several subdirs in the main directory. Each dirname follows the format that 
    `description-formula-type`, for example, `water-H2O-molecule`, is a system with structures 
    that have one single water molecule.

    """

    def __init__(
        self,
        dataset_path: Union[str, pathlib.Path] = "./",
        batchsize: int = 32,
        train_ratio: float = 0.9,
        random_seed: int = None,
        prop_keys: List[List[str]] = DEFAULT_PROP_MAP_KEYS,
        *args,
        **kwargs,
    ) -> None:
        """"""
        super().__init__(directory=dataset_path, random_seed=random_seed)

        self.batchsize = batchsize
        self.train_ratio = train_ratio

        self.prop_keys = prop_keys

        return

    def load(self) -> List[pathlib.Path]:
        """Load dataset.

        All directories that have xyz files in `self.directory`.

        TODO:
            * Other file formats.

        """
        data_dirs = traverse_xyzdirs(self.directory)
        data_dirs = sorted(data_dirs)

        return data_dirs

    def load_frames(self, *args, **kwargs):
        """"""
        data_dirs = traverse_xyzdirs(self.directory)
        data_dirs = sorted(data_dirs)

        names = [
            "+".join(str(x.relative_to(self.directory)).split("/")) for x in data_dirs
        ]

        nframes_tot, frames_list = 0, []
        for i, p in enumerate(data_dirs):
            curr_frames = []
            xyzpaths = sorted(list(p.glob("*.xyz")))
            for x in xyzpaths:
                curr_frames.extend(read(x, ":"))
            curr_nframes = len(curr_frames)
            nframes_tot += curr_nframes
            self._debug(f"{i:>4d} {str(p)} -> {len(curr_frames)}")
            frames_list.append(curr_frames)
        self._debug(f"Number of frames: {nframes_tot}")

        pairs = []
        for n, x in zip(names, frames_list):
            pairs.append([n, x])

        # - map keys
        should_map_keys = False
        for mapping_pairs in self.prop_keys:
            dst_key, src_key = mapping_pairs
            if dst_key != src_key:
                should_map_keys = True
            else:
                ...
        else:
            ...
        if should_map_keys:
            for n, x in pairs:
                for a in x:
                    map_atoms_data(a, self.prop_keys)

        return pairs

    def split_train_and_test(
        self,
    ):
        """Read structures and split them into train and test."""
        self._print("--- auto data reader ---")
        data_dirs = self.load()
        self._debug(data_dirs)

        # - aggregate data folders
        # dir_indices = list(range(data_dirs))
        system_paths = []
        for d in data_dirs:
            if is_a_valid_system_name(d.name):
                system_paths.append(d)
            else:
                system_paths.append(d.parent)

        system_groups = {}
        for k, v in itertools.groupby(enumerate(system_paths), key=lambda x: str(x[1])):
            if k not in system_groups:
                system_groups[k] = [data_dirs[e[0]] for e in v]
            else:
                system_groups[k].extend([data_dirs[e[0]] for e in v])

        # -- convert to tuple
        system_groups_ = []
        for k, v in system_groups.items():
            system_groups_.append([k, v])
        system_groups = system_groups_

        # -
        batchsizes = self.batchsize
        nsystems = len(system_groups)
        if isinstance(batchsizes, int):
            batchsizes = [batchsizes] * nsystems
        assert (
            len(batchsizes) == nsystems
        ), "Number of systems and batchsizes are inconsistent."

        # read configurations
        set_names = []
        train_size, test_size = [], []
        train_frames, test_frames = [], []
        adjusted_batchsizes = []  # auto-adjust batchsize based on nframes
        for i, (curr_system_group, curr_batchsize) in enumerate(
            zip(system_groups, batchsizes)
        ):
            curr_system = pathlib.Path(curr_system_group[0])
            set_name = "+".join(str(curr_system.relative_to(self.directory)).split("/"))
            set_names.append(set_name)
            self._print(f"System {set_name} Batchsize {curr_batchsize}")
            frames = []  # all frames in this subsystem
            for curr_subsystem in curr_system_group[1]:
                self._print(f"  {curr_subsystem.relative_to(curr_system)}")
                xyz_fpaths = list(curr_subsystem.glob("*.xyz"))
                xyz_fpaths.sort()  # sort by alphabet
                for p in xyz_fpaths:
                    # read and split dataset
                    p_frames = read(p, ":")
                    p_nframes = len(p_frames)
                    frames.extend(p_frames)
                    self._print(f"    subsystem: {p.name} number {p_nframes}")

            # split dataset and get adjusted batchsize
            # TODO: adjust batchsize of train and test separately
            nframes = len(frames)
            if nframes <= curr_batchsize:
                if nframes == 1 or curr_batchsize == 1:
                    new_batchsize = 1
                else:
                    new_batchsize = int(2 ** np.floor(np.log2(nframes)))
                adjusted_batchsizes.append(new_batchsize)
                # NOTE: use same train and test set
                #       since they are very important structures...
                train_index = list(range(nframes))
                test_index = list(range(nframes))
            else:
                if nframes == 1 or curr_batchsize == 1:
                    new_batchsize = 1
                    train_index = list(range(nframes))
                    test_index = list(range(nframes))
                else:
                    new_batchsize = curr_batchsize
                    # - assure there is at least one batch for test
                    #          and number of train frames is integer times of batchsize
                    ntrain = int(
                        np.floor(nframes * self.train_ratio / new_batchsize)
                        * new_batchsize
                    )
                    train_index = self.rng.choice(nframes, ntrain, replace=False)
                    test_index = [x for x in range(nframes) if x not in train_index]
                adjusted_batchsizes.append(new_batchsize)

            ntrain, ntest = len(train_index), len(test_index)
            train_size.append(ntrain)
            test_size.append(ntest)

            self._print(
                f"    ntrain: {ntrain} ntest: {ntest} ntotal: {nframes} batchsize: {new_batchsize}"
            )

            curr_train_frames = [frames[train_i] for train_i in train_index]
            curr_test_frames = [frames[test_i] for test_i in test_index]

            # train
            train_frames.append(curr_train_frames)
            n_train_frames = sum([len(x) for x in train_frames])

            # test
            test_frames.append(curr_test_frames)
            n_test_frames = sum([len(x) for x in test_frames])
            self._print(
                f"  Current Dataset -> ntrain: {n_train_frames} ntest: {n_test_frames}"
            )

        assert len(train_size) == len(
            test_size
        ), "inconsistent train_size and test_size"
        train_size = sum(train_size)
        test_size = sum(test_size)
        self._print(f"Total Dataset -> ntrain: {train_size} ntest: {test_size}")

        # - map keys
        should_map_keys = False
        for mapping_pairs in self.prop_keys:
            dst_key, src_key = mapping_pairs
            if dst_key != src_key:
                should_map_keys = True
            else:
                ...
        else:
            ...

        if should_map_keys:
            for curr_frames in train_frames:
                for a in curr_frames:
                    map_atoms_data(a, self.prop_keys)
            for curr_frames in test_frames:
                for a in curr_frames:
                    map_atoms_data(a, self.prop_keys)

        return set_names, train_frames, test_frames, adjusted_batchsizes

    def transfer(self, frames: List[Atoms]):
        """Add structures into the dataset."""
        # - check chemical symbols
        system_dict = {}  # {formula: [indices]}

        formulae = [a.get_chemical_formula() for a in frames]
        for k, v in itertools.groupby(enumerate(formulae), key=lambda x: x[1]):
            system_dict[k] = [x[0] for x in v]

        # - transfer data
        for formula, curr_indices in system_dict.items():
            # -- TODO: check system type
            system_type = self.system  # currently, use user input one
            # -- name = description+formula+system_type
            dirname = "-".join([self.directory.parent.name, formula, system_type])
            target_subdir = self.target_dir / dirname
            target_subdir.mkdir(parents=True, exist_ok=True)

            # -- save frames
            curr_frames = [frames[i] for i in curr_indices]
            curr_nframes = len(curr_frames)

            strname = self.version + ".xyz"
            target_destination = self.target_dir / dirname / strname
            if not target_destination.exists():
                write(target_destination, curr_frames)
                self._print(f"nframes {curr_nframes} -> {target_destination.name}")
            else:
                warnings.warn(f"{target_destination} exists.", UserWarning)

        return

    def as_dict(self):
        """"""
        dataset_params = {}
        dataset_params["name"] = self.name
        dataset_params["dataset_path"] = str(self.directory.resolve())
        dataset_params["batchsize"] = self.batchsize
        dataset_params["train_ratio"] = self.train_ratio

        dataset_params = copy.deepcopy(dataset_params)

        return dataset_params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import pathlib

from ..core.register import registers

from .correction import correct
registers.operation.register(correct)

from .convert import convert_dataset



if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import itertools
from typing import List

from ase import Atoms
from ase.io import read, write
from ase.calculators.singlepoint import SinglePointCalculator

from ..core.operation import Operation


def merge_results(host_frames: List[Atoms], corr_frames: List[Atoms]):
    """"""
    frames = []
    for host, corr in zip(host_frames, corr_frames):
        # TODO: check atoms consistent?
        atoms = Atoms(
            symbols=copy.deepcopy(host.get_chemical_symbols()),
            positions=copy.deepcopy(host.get_positions()),
            cell=copy.deepcopy(host.get_cell(complete=True)),
            pbc=copy.deepcopy(host.get_pbc()),
            tags = host.get_tags() # retain this for molecules
        )
        if host.get_kinetic_energy() > 0.: # retain this for MD
            atoms.set_momenta(host.get_momenta()) 
        # TODO: add info and arrays?
        calc = SinglePointCalculator(
            atoms, 
            energy = host.get_potential_energy() + corr.get_potential_energy(),
            forces = host.get_forces(apply_constraint=True) + corr.get_forces(apply_constraint=True),
        )
        atoms.calc = calc
        frames.append(atoms)

    return frames


class correct(Operation):

    """"""

    def __init__(self, structures, computer, directory="./") -> None:
        """"""
        input_nodes = [structures, computer]
        super().__init__(input_nodes, directory)

        return
    
    def forward(self, structures, computer):
        """"""
        super().forward()

        computer = computer[0] # TODO: accept several computers...
        computer._share_wdir = True

        # - structures should be a Tempdata node for now
        computer_status = [False for i in range(len(structures))]
        for i, (curr_name, curr_frames) in enumerate(structures):
            nframes = len(curr_frames)
            print(curr_name, len(curr_frames))
            computer.directory = self.directory/curr_name
            computer.batchsize = nframes
            computer.run(curr_frames)
            computer.inspect(resubmit=True)
            if (computer.get_number_of_running_jobs() == 0):
                computer_status[i] = True
                continue
        
        if all(computer_status):
            self._print("correction computation finished...")
            self.status = "finished"
            new_structures = []
            for i, (curr_name, curr_frames) in enumerate(structures):
                computer.directory = self.directory/curr_name
                # TODO: neew a better unified interface
                curr_corr_frames = computer.retrieve()
                if not computer._share_wdir:
                    curr_corr_frames = itertools.chain(*curr_corr_frames)
                curr_new_frames = merge_results(curr_frames, curr_corr_frames)
                write(self.directory/f"{curr_name}"/"merged.xyz", curr_new_frames)
                new_structures.append([curr_name, curr_new_frames])

        return new_structures


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import dataclasses
import pathlib
import time

from typing import Union, List

from .. import registers

from ..expedition import AbstractExpedition


class SimulatedAnnealing(AbstractExpedition):

    name: str = "simulated_annealing"

    #: Name of the folder that stores all worker computations.
    comput_dirname: str = "tmp_folder"

    def __init__(
        self,
        builder,
        temperatures: List[float],
        directory: Union[str, pathlib.Path] = "./",
        random_seed: Union[int, dict] = None,
        *args,
        **kwargs,
    ):
        super().__init__(directory, random_seed, *args, **kwargs)

        # - check system type
        if isinstance(builder, dict):
            builder_params = copy.deepcopy(builder)
            builder_method = builder_params.pop("method")
            builder = registers.create(
                "builder", builder_method, convert_name=False, **builder_params
            )
        else:
            builder = builder
        self.builder = builder

        # -
        self.temperatures = temperatures

        self.num_slices = len(self.temperatures)

        return

    def run(self, *args, **kwargs):
        """"""
        super().run()

        # ---
        structures = self.builder.run()
        nstructures = len(structures)
        self._print(f"Number of input structures: {nstructures}.")

        num_slices = len(self.temperatures)

        # ---
        rng_states = []
        for istep in range(num_slices):
            curr_temperature = self.temperatures[istep]
            if istep > 0:
                # - get each candidates' final rng_states
                prev_structures, prev_rng_states = [], []
                for icand in range(nstructures):
                    curr_wdir = (
                        self.directory
                        / self.comput_dirname
                        / f"gen{istep-1}"
                        / f"cand{icand}"
                    )
                    ckpt_wdir = self.worker.driver._find_latest_checkpoint(curr_wdir)
                    prev_atoms, prev_rng_state = self.worker.driver._load_checkpoint(
                        ckpt_wdir
                    )
                    self._print(f"{prev_atoms =}")
                    self._print(f"{prev_rng_state =}")
                    prev_structures.append(prev_atoms)
                    prev_rng_states.append(prev_rng_state)
                    # --
                    if hasattr(self.worker.driver.calc, "calcs"):
                        for calc in self.worker.driver.calc.calcs:
                            if hasattr(calc, "_load_checkpoint"):
                                calc._load_checkpoint(
                                    ckpt_wdir,
                                    dst_wdir=self.directory
                                    / self.comput_dirname
                                    / f"gen{istep}"
                                    / f"cand{icand}",
                                    start_step=prev_atoms.info["step"],
                                )
                structures = prev_structures
                rng_states = prev_rng_states
            else:
                ...
            step_converged = self._irun(istep, structures, curr_temperature, rng_states)
            if not step_converged:
                self._print(f"Wait Slice {istep} to finish.")
                break
            else:
                ...
        else:
            self._print("SlicedExpedition is converged.")
            with open(self.directory / "FINISHED", "w") as fopen:
                fopen.write(
                    f"FINISHED AT {time.asctime( time.localtime(time.time()) )}."
                )

        return

    def _irun(self, istep: int, structures, temperature: float, rng_states: list):
        """"""
        self._print(f"===== SlicedExpedition Step {istep} =====")
        self._print(f"{temperature =}")

        # - We need a copy of self.worker as we may change some
        #   of the driver's settings.
        worker = self._make_step_worker(istep=istep)

        # - update driver...
        worker.driver.setting.update(temp=temperature)

        # - We need copy some files to continue exploration...
        if istep > 0:
            if hasattr(worker.driver.calc, "calcs"):
                self._print(f"{worker.driver.calc =}")
                for calc in worker.driver.calc.calcs:
                    if hasattr(calc, "_load_checkpoint"):
                        ...
            else:
                ...
        else:
            ...

        # -
        # TODO: NonLocal Scheduler will not save rng_states...
        worker.run(structures, rng_states=rng_states)
        worker.inspect(resubmit=True)
        if worker.get_number_of_running_jobs() == 0:
            step_converged = True
        else:
            step_converged = False

        return step_converged

    def _make_step_worker(self, istep: int):
        """"""
        if hasattr(self.worker.potter, "remove_loaded_models"):
            self.worker.potter.remove_loaded_models()
        worker = copy.deepcopy(self.worker)

        worker.directory = self.directory / self.comput_dirname / f"gen{istep}"

        return worker

    def read_convergence(self) -> bool:
        """"""
        converged = False
        if (self.directory / "FINISHED").exists():
            converged = True

        return converged

    def get_workers(self):
        """"""
        if hasattr(self.worker.potter, "remove_loaded_models"):
            self.worker.potter.remove_loaded_models()

        workers = []
        for istep in range(self.num_slices):
            curr_worker = copy.deepcopy(self.worker)
            curr_worker.directory = self.directory / self.comput_dirname / f"gen{istep}"
            workers.append(curr_worker)

        return workers


if __name__ == "__main__":
    ...
